
# =====================================================================================================================
# **NOTICE**
# THIS IS A GENERATED FILE USING THE SOURCE FILE:
#  crab\common\lib\jvm\seeq-names\src\main\java\com\seeq\common\seeqnames\SeeqNames.java
# =====================================================================================================================

class SeeqNames:
    class FormulaType:
        scalar = 'Scalar'
        signal = 'Signal'
        sample = 'Sample'
        condition = 'Condition'
        capsule = 'Capsule'
        formula_item = 'FormulaItem'
        onnx_prediction_model = 'OnnxPredictionModel'
        onnx_anomaly_model = 'OnnxAnomalyModel'
    
    class AuditTrail:
        configuration_option = 'ConfigurationOption'
    
    class ImpactReport:
        impact_table_name = 'VCC_ImpactTable_v1'
        impact = 'impact'
        impact_name = 'Name'
        impact_description = 'Description'
        impact_owner = 'Impact Owner'
        impact_status = 'Status'
        impact_formula = 'Formula'
        impact_savings_type = 'Savings Type'
        impact_frequency = 'Frequency'
        impact_unit = 'Unit'
        impact_asset = 'Asset'
        impact_start_date = 'Impact Start Date'
        impact_end_date = 'Impact End Date'
        impact_quantify_amount = 'Impact Quantify Amount'
        assigned_to_worksheet = 'Worksheet Id'
        assigned_to_workbook = 'Workbook Id'
        impact_site = 'Impact Site'
        manager = 'Manager'
        impact_lessons_learned = 'Lesson Learned'
        impact_formula_param = 'Formula Param'
        impact_based_on = 'Based on'
    
    class GraphQL:
        class DataLoaderRegistry:
            creator_metadata_loader = 'CREATOR_METADATA_LOADER'
            item_metadata_loader = 'ITEM_METADATA_LOADER'
        
    
    class SeeqContextNameSpace:
        seeq_label_namespace = '__'
        usage_restricted_label_namespace = '__Seeq.'
        vantage_label_category_namespace = '__Vantage'
    
    class MlService:
        class Clustering:
            class Connection:
                name = 'Seeq ML Clustering'
                url = 'cluster'
                udf_package = 'seeqClustering'
                udf_name = 'cluster'
            
        
        class IsolationForest:
            class Connection:
                name = 'Seeq ML Isolation Forest'
                url = 'if'
                udf_package = 'seeqIF'
                udf_name = 'isolationForest'
            
        
        class SelfOrganizingMap:
            class Connection:
                name = 'Seeq ML Self-Organizing Map'
                url = 'som'
                udf_package = 'seeqSOM'
                udf_name = 'selfOrganizingMap'
            
        
        class PlsRegression:
            class Connection:
                name = 'Seeq ML PLS Regression'
                url = 'pls'
                udf_package = 'seeqPLS'
                udf_name = 'plsRegression'
            
        
    
    class Consumption:
        screenshot = 'Screenshot'
        topic = 'Topic'
    
    class Capabilities:
        class ScalingTables:
            data_id = 'ScalingTablesCapability'
            collective_name = 'scaling tables users'
        
        class LogViewer:
            data_id = 'LogViewerCapability'
            collective_name = 'log viewers'
        
        class DatasourceAdministration:
            data_id = 'DatasourceAdministrationCapability'
            collective_name = 'datasource administrators'
        
        class UserAdministration:
            data_id = 'UserAdministrationCapability'
            collective_name = 'user administrators'
        
        class AuditTrail:
            data_id = 'AuditTrailCapability'
            collective_name = 'audit trail viewers'
        
        class AnalyticsAdministration:
            data_id = 'AnalyticsAdministrationCapability'
            collective_name = 'analytics administrators'
        
        class Administration:
            data_id = 'Administrators'
            collective_name = 'administrators'
        
    
    class RequestIdPrefixes:
        unique = 'R^'
        screenshot = 'ScreenshotCapture^'
        invocation = 'Invocation^'
    
    class PdfSettings:
        class MarginUnits:
            inches = 'in'
            millimeters = 'mm'
            centimeters = 'cm'
            pixels = 'px'
        
        class PageSize:
            letter = 'Letter'
            legal = 'Legal'
            a5 = 'A5'
            a4 = 'A4'
            a3 = 'A3'
            tabloid = 'Tabloid'
        
        class PageOrientation:
            portrait = 'Portrait'
            landscape = 'Landscape'
        
        class Defaults:
            page_orientation = 'Portrait'
            page_size = 'Letter'
            page_margin = '0.5in'
        
    
    class Database:
        class Utils:
            empty_guid = '00000000-0000-0000-0000-000000000000'
        
        class Schemas:
            materialized_tables = 'materialized_tables'
        
    
    class MaterializedTables:
        datum_id_column = 'datum id'
        item_id_column = 'item id'
        event_data_column = 'event data'
        context_table_prefix = 'context_'
        created_at_column = 'created at'
        unique_row_id = 'Existing Row Identifier'
        csv_item_id_column = 'csv item id'
        class Parameters:
            property_name = 'propertyName'
            property_names = 'propertyNames'
            property_value = 'propertyValue'
            property_match_operator = 'propertyMatchOperator'
            property_match_operators = 'propertyMatchOperators'
            level = 'level'
            separator = 'separator'
            formula = 'formula'
            exclude_globally_scoped = 'excludeGloballyScoped'
            parameters_name = 'parameters'
            name = 'name'
            description = 'description'
            column_index_data_id = 'columnIndexDataId'
            scoped_to = 'scopedTo'
            identity_id = 'identityId'
            permission_string = 'permissionString'
            security_string = 'securityString'
            column_index = 'columnIndex'
            column_indexes = 'columnIndexes'
            column_index_item = 'columnIndexItem'
            column_index_value = 'columnIndexValue'
            column_index_source_text = 'columnIndexSourceText'
            column_index_replacement = 'columnIndexReplacement'
            column_index_unit_of_measure = 'columnIndexUnitOfMeasure'
            constant = 'constant'
            is_root = 'isRoot'
            item_type = 'itemType'
            item_types = 'itemTypes'
            regex = 'regex'
            replacement = 'replacement'
            remove_non_matches = 'removeNonMatches'
            lookup_value = 'lookupValue'
            variable_parameters = 'variableParameters'
            variable_parameter_strings = 'variableParameterStrings'
            prefix = 'prefix'
            suffix = 'suffix'
        
        class Rules:
            ancestor = 'ancestor'
            asset_creator = 'assetCreator'
            concat_columns = 'concatColumns'
            descendant = 'descendant'
            event_property = 'eventProperty'
            formula_creator = 'formulaCreator'
            get_item_property = 'getItemProperty'
            item_search = 'itemSearch'
            path = 'path'
            path_search = 'pathSearch'
            numeric_scalar_creator = 'numericScalarCreator'
            set_item_property = 'setItemProperty'
            string_scalar_creator = 'stringScalarCreator'
            text_extractor = 'textExtractor'
            text_replacement = 'textReplacement'
            tree_path_creator = 'treePathCreator'
            manager = 'manager'
            item_type = 'itemType'
            datasource = 'datasource'
            context_condition = 'contextCondition'
            class ScalarCreator:
                base_scalar_creator = 'scalarCreator'
                numeric_scalar_creator = 'numericScalarCreator'
                string_scalar_creator = 'stringScalarCreator'
            
            class Overrides:
                override = 'Override'
                boolean_override = 'booleanOverride'
                numeric_override = 'numericOverride'
                string_override = 'stringOverride'
                timestamp_override = 'timestampOverride'
                u_u_i_d_override = 'uuidOverride'
            
            class Constant:
                base_constant = 'constant'
                boolean_constant = 'booleanConstant'
                numeric_constant = 'numericConstant'
                string_constant = 'stringConstant'
                timestamp_constant = 'timestampConstant'
                u_u_i_d_constant = 'uuidConstant'
                boolean_lookup_constant = 'booleanLookupConstant'
                numeric_lookup_constant = 'numericLookupConstant'
                string_lookup_constant = 'stringLookupConstant'
                timestamp_lookup_constant = 'timestampLookupConstant'
                u_u_i_d_lookup_constant = 'uuidLookupConstant'
            
        
        class FormulaAcross:
            asset_column = 'Asset'
            path_column = 'Path'
            datasource_column = 'Datasource'
            value_uom_column = 'Value UOM'
        
    
    class Injection:
        system_state = 'SystemState'
        system_user = 'SystemUser'
        tree_definition = 'TreeDefinition'
    
    class Ignition:
        agent_name = 'Ignition Module Agent'
    
    class Connectors:
        seeq_developer_name = 'Seeq'
        unknown_third_party_developer_name = 'third-party developer'
        connections_json_key = 'Connections'
        datasource_managed_json_key = 'DatasourceManaged'
        class Connections:
            name_json_key = 'Name'
            datasource_id_json_key = 'Id'
            enabled_json_key = 'Enabled'
            indexing_json_key = 'Indexing'
            frequency_json_key = 'Frequency'
            next_indexing_at_json_key = 'Next'
            on_startup_and_config_change_json_key = 'OnStartupAndConfigChange'
            transforms_json_key = 'Transforms'
            max_concurrent_requests_json_key = 'MaxConcurrentRequests'
            max_results_per_request_json_key = 'MaxResultsPerRequest'
            field_configurations_json_key = 'FieldConfigurations'
            class SyncStatus:
                unknown = 'SYNC_UNKNOWN'
                initializing = 'SYNC_INITIALIZING'
                in_progress = 'SYNC_IN_PROGRESS'
                archiving_deleted_items = 'SYNC_ARCHIVING_DELETED_ITEMS'
                complete = 'SYNC_COMPLETE'
                success = 'SYNC_SUCCESS'
                failed = 'SYNC_FAILED'
            
            class Status:
                connecting = 'CONNECTING'
                connected = 'CONNECTED'
                disconnected = 'DISCONNECTED'
                disabled = 'DISABLED'
            
            class SyncMode:
                none = 'NONE'
                incremental = 'INCREMENTAL'
                full = 'FULL'
            
            class SyncResult:
                in_progress = 'IN_PROGRESS'
                success = 'SUCCESS'
                failed = 'FAILED'
                interrupted = 'INTERRUPTED'
            
            class MultivaluedSamplesBehavior:
                keep_last = 'Keep Last'
                stagger = 'Stagger'
            
        
    
    class Agents:
        agent_keys_folder_name = 'agent-keys'
        connectors_json_key = 'Connectors'
        class WebSockets:
            ping_prefix = 'SEEQ-WS-PING-MSG-'
        
        class Status:
            connecting = 'CONNECTING'
            connected = 'CONNECTED'
            disconnected = 'DISCONNECTED'
            pre_provisioned = 'PREPROVISIONED'
        
        class Connectors:
            name_json_key = 'Name'
            enabled_json_key = 'Enabled'
        
    
    class TreeType:
        seeq_workbench = 'Seeq Workbench'
    
    class OData:
        export_entity_set_suffix = '_DataSet'
        o_data_v4_service_suffix = '.svc/'
    
    class CleanUpJobsGroup:
        annotation_images = 'AnnotationOrphanedImageCleanerJob'
        worksheet_maintenance = 'WorksheetMaintenanceJob'
    
    class JobGroup:
        screenshot = 'ScreenshotCapture'
        date_range = 'DateRangeJobs'
        report = 'ReportJobs'
        notebook = 'NotebookJobs'
        clean_up = 'CleanUpJobs'
        report_template = 'ReportTemplateJobs'
        condition_monitor = 'ConditionMonitorNotifications'
        worksheet_thumbnail = 'WorksheetThumbnail'
        item_finder = 'ItemFinder'
    
    class Logging:
        migrations_complete_file = 'migrations_complete_'
        class RemoteAgents:
            m_d_c_filename_key = 'agentlogfilename'
        
        class Appserver:
            m_d_c_filename_key = 'logfilename'
            default_file = 'appserver'
            startup_file = 'appserver-startup'
            upgrade_file = 'appserver-upgrade'
            async_upgrade_file = 'appserver-async-upgrade'
            licensing_file = 'appserver-licensing'
            requests = 'appserver-requests'
            request_monitors = 'appserver-requestmonitors'
            units = 'appserver-units'
            users = 'appserver-users'
            class IntervalAnalysis:
                fetch_critic = 'appserver-intervalanalysis-fetchcritic'
            
        
    
    class RemoteDatasourceClasses:
        l_d_a_p = 'LDAP'
        o_auth2 = 'OAuth 2.0'
        windows_auth = 'Windows Auth'
        a_d_x = 'ADX'
        a_p_r_m = 'AspenTech APRM'
        big_query = 'BigQuery'
        chaos_monkey = 'Chaos Monkey'
        cognite = 'Cognite'
        c_o_n_n_e_c_t_data_services = 'CONNECT Data Services'
        cyg_net = 'CygNet'
        example_data = 'Time Series CSV Files'
        add_on_calc_python = 'Add-on Calculation Python'
        add_on_calc_webhook = 'Add-on Calculation Webhook'
        add_on_calc_excel = 'Add-on Calculation Excel VBA'
        file_folders = 'FileFolders'
        g_e_proficy = 'GE Proficy'
        g_p_m = 'GreenPowerMonitor'
        i_b_a_h_d = 'IbaHD'
        ignition = 'Ignition Gateway'
        influx_d_b = 'InfluxDB'
        influx_d_b2 = 'InfluxDB 2'
        i_p21 = 'AspenTech IP21'
        met_net = 'MetNet'
        n_o_a_a_weather = 'NOAA Weather Service'
        omnia = 'Equinor Omnia Plant'
        o_p_c_h_d_a = 'OPC-HDA'
        o_p_c_u_a = 'OPC-UA'
        o_s_isoft_a_f = 'OSIsoft AF'
        o_s_isoft_a_f_event_frames = 'OSIsoft AF Event Frames'
        o_s_isoft_p_i = 'OSIsoft PI'
        seeq_to_seeq = 'Seeq to Seeq'
        s_q_l = 'SQL'
        stress_test = 'Stress Test'
        time_stream = 'TimeStream'
        tree_file = 'Tree File'
        w_i_t_s_m_l = 'WITSML'
        wonderware = 'Wonderware'
        x_h_q = 'XHQ'
    
    class SeeqSaasAdminDatasource:
        datasource_class = 'OAuth 2.0'
        datasource_id = 'Seeq Saas Admin'
        datasource_name = 'Seeq Saas Admin'
    
    class IdentityProviderDatasourceClasses:
        l_d_a_p = 'LDAP'
        o_auth2 = 'OAuth 2.0'
        seeq_auth = 'Auth'
        windows_auth = 'Windows Auth'
    
    class LocalDatasources:
        class MaterializedTables:
            datasource_class = 'Seeq Materialized Tables'
        
        class SeeqDataLab:
            datasource_class = 'Seeq Data Lab'
        
        class SeeqMetadataItemStorage:
            datasource_class = 'Seeq Metadata'
            datasource_id = 'Seeq Metadata'
            datasource_name = 'Seeq Metadata'
        
        class SeeqCalculationItemStorage:
            datasource_class = 'Seeq Calculations'
            datasource_id = 'Seeq Calculations'
            datasource_name = 'Seeq Calculations'
        
        class Monitors:
            datasource_class = 'Monitors'
            datasource_id = 'monitors'
            datasource_name = 'Seeq Monitors'
        
        class Authentication:
            datasource_class = 'Auth'
            datasource_id = 'Seeq'
            datasource_name = 'Seeq'
            password_special_characters = '!@#$%^&*()_+-=[]{}|;:,.<>?/'
        
        class FileSignalStorage:
            datasource_class = 'Seeq - Signal Datasource'
            datasource_id = 'Signal Datasource'
            datasource_name = 'Seeq Signal Datasource'
        
        class ContextDatumStorage:
            datasource_class = 'Context Datum Storage'
            datasource_id = 'Context Datum Storage'
            datasource_name = 'Context Datum Storage'
        
        class PostgresDatums:
            datasource_class = 'Postgres Datums'
            datasource_id = 'default'
            datasource_name = 'Seeq Conditions Database'
        
        class Datafile:
            datasource_class = 'Imported CSV Files'
            datasource_id = 'Imported CSV Files'
            datasource_name = 'Imported CSV Files'
        
    
    class SwaggerApiNames:
        agents = 'Agents'
        add_ons = 'AddOns'
        audit = 'Audit'
        auth = 'Auth'
        assets = 'Assets'
        folders = 'Folders'
        networks = 'Networks'
        items = 'Items'
        reports = 'Reports'
        jobs = 'Jobs'
        logs = 'Logs'
        trees = 'Trees'
        sample_series = 'SampleSeries'
        signals = 'Signals'
        capsules = 'Capsules'
        conditions = 'Conditions'
        metrics = 'Metrics'
        formulas = 'Formulas'
        functions = 'Functions'
        requests = 'Requests'
        tables = 'Tables'
        workbooks = 'Workbooks'
        worksheets = 'Worksheets'
        export = 'Export'
        annotations = 'Annotations'
        datasources = 'Datasources'
        datafiles = 'Datafiles'
        users = 'Users'
        user_groups = 'UserGroups'
        access_keys = 'AccessKeys'
        usage = 'Usage'
        scalars = 'Scalars'
        subscriptions = 'Subscriptions'
        system = 'System'
        import_ = 'Import'
        monitors = 'Monitors'
        projects = 'Projects'
        content = 'Content'
        report_templates = 'ReportTemplates'
        condition_monitors = 'ConditionMonitors'
        plugins = 'Plugins'
        displays = 'Displays'
        display_templates = 'DisplayTemplates'
        notification_configurations = 'NotificationConfigurations'
        label = 'Label'
        context = 'Context'
        table_definitions = 'TableDefinitions'
        models = 'Models'
        table_overrides = 'TableOverrides'
        graph_q_l = 'GraphQL'
        unstable = 'Unstable'
        scim = 'SCIM'
    
    class Channels:
        agents_status = 'agents-status'
        datasources_status = 'datasources-status'
        broadcast = 'broadcast'
        live_screenshot = 'live-screenshot'
        requests_progress = 'requests-progress'
        workbook_channel = 'workbook'
        locking_channel = 'workbook-locking'
        thumbnail_channel = 'workbook-thumbnail'
        workstep_channel = 'workstep'
        report_update_channel = 'report-update'
        table_materializer_channel = 'table-materializer'
    
    class API:
        ace = '/ace'
        add_ons = '/add-ons'
        add_on_tools = '/addOnTools'
        agents = '/agents'
        audit = '/audit'
        auth = '/auth'
        assets = '/assets'
        gen_a_i = '/genai'
        folders = '/folders'
        networks = '/networks'
        items = '/items'
        item_finders = '/finders'
        reports = '/reports'
        jobs = '/jobs'
        logs = '/logs'
        batch = '/batch'
        trees = '/trees'
        asset_trees = '/trees/assets'
        signals = '/signals'
        samples = '/samples'
        plot_samples = '/plot-samples'
        capsules = '/capsules'
        conditions = '/conditions'
        condition_monitors = '/condition-monitors'
        metrics = '/metrics'
        thresholds = '/thresholds'
        formulas = '/formulas'
        docs = '/docs'
        functions = '/functions'
        packages = '/packages'
        scalars = '/scalars'
        properties = '/properties'
        requests = '/requests'
        tables = '/tables'
        workbooks = '/workbooks'
        worksheets = '/worksheets'
        worksteps = '/worksteps'
        export = '/export'
        annotations = '/annotations'
        datasources = '/datasources'
        datafiles = '/datafiles'
        users = '/users'
        user_groups = '/usergroups'
        access_keys = '/accesskeys'
        o_data = '/odata.svc'
        o_data_v4 = '/odata'
        system = '/system'
        import_ = '/import'
        status = '/status'
        shutdown = '/shutdown'
        screenshots = '/screenshots'
        subscriptions = '/subscriptions'
        monitors = '/system/monitors'
        configuration = '/system/configuration'
        installer = '/system/installer'
        projects = '/projects'
        content = '/content'
        report_templates = '/report-templates'
        date_ranges = '/date-ranges'
        asset_selections = '/asset-selections'
        plugins = '/plugins'
        displays = '/displays'
        display_templates = '/display-templates'
        usage = '/usage'
        notification_configurations = '/notification-configurations'
        labels = '/labels'
        context = '/context'
        context_comments = '/comments'
        context_labels = '/labels'
        context_opaque = '/opaque'
        context_numeric = '/numeric'
        table_definitions = '/table-definitions'
        models = '/models'
        graph_q_l = '/graphql'
        cache = '/cache'
        scim_users = 'Users'
        scim_groups = 'Groups'
        scim_token = 'token'
        class ErrorMessages:
            attempted_to_set_scope_on_a_globally_scoped_item = 'Attempted to set scope on a globally scoped item'
            cancelled_due_to_too_many_retries = 'The request was canceled due to being retried too many times'
        
        class AddOnToolLinkType:
            window = 'window'
            tab = 'tab'
            none = 'none'
        
        class AddOnToolLaunchLocation:
            homescreen = 'homescreen'
            tools_pane = 'toolsPane'
        
        class PropertyIds:
            request_information = 'com.seeq.requestinformation'
        
        class Cookies:
            auth = 'sq-auth'
            csrf = 'sq-csrf'
        
        class QueryParams:
            admin = 'admin'
            client_interactive_session_id = 'interactiveId'
            creator = 'creator'
            name = 'name'
            request_id = 'requestId'
            content_count = 'contentCount'
        
        class Headers:
            accept_ranges = 'Accept-Ranges'
            async_ = 'x-sq-async'
            async_job = 'x-sq-async-job'
            auth = 'x-sq-auth'
            identity_path = 'x-sq-identity-path'
            content_range = 'Content-Range'
            csrf = 'x-sq-csrf'
            forwarded_for = 'x-forwarded-for'
            interactive_session_id = 'Session-Id'
            pool_id = 'x-sq-pool-id'
            pool_connected_on_degraded_state = 'x-sq-pool-connected-on-degraded-state'
            range = 'Range'
            request_id = 'x-sq-request-id'
            request_origin = 'x-sq-origin'
            request_origin_u_r_l = 'x-sq-origin-url'
            request_origin_label = 'x-sq-origin-label'
            server = 'Server'
            server_meters = 'Server-Meters'
            server_timing = 'Server-Timing'
            expect = 'Expect'
            date_range_start = 'x-sq-daterange-start'
            date_range_end = 'x-sq-daterange-end'
            class Timings:
                datasource = 'Datasource'
                cache = 'Cache'
                processing = 'Processing'
                g_c = 'GC'
                request_queue = 'Request Queue'
                calc_engine_queue = 'Calc Engine Queue'
                seeq_database = 'Seeq Database'
                total = 'Total'
            
            class Meters:
                datasource_samples_read = 'Datasource Samples Read'
                datasource_capsules_read = 'Datasource Capsules Read'
                cache_samples_read = 'Cache Samples Read'
                cache_capsules_read = 'Cache Capsules Read'
                cache_in_memory_samples_read = 'Cache In-Memory Samples Read'
                cache_in_memory_capsules_read = 'Cache In-Memory Capsules Read'
                database_items_read = 'Database Items Read'
                database_relationships_read = 'Database Relationships Read'
            
        
        class Flags:
            all_properties = '@allProperties'
            exclude_globally_scoped = '@excludeGloballyScoped'
            include_unsearchable = '@includeUnsearchable'
        
        class TranslationKeyPrefixes:
            capabilities = 'API.CAPABILITIES.'
        
    
    class UnitTest:
        timestamp1 = '1990-03-07T09:58:20.888888Z'
        timestamp2 = '1990-04-08T10:59:21.999999Z'
        non_existent_guid = '1CB64296-FC3E-4221-9A7F-947A0AB99807'
    
    class Edges:
        owns = 'owns'
        created = 'created'
        installed = 'installed'
        contains = 'contains'
        folder_contains = 'folder contains'
        member_of = 'member of'
        interested_in = 'interested in'
        reports_on = 'reports on'
        replies_to = 'replies to'
        recently_accessed = 'recently accessed'
        has_worksheet = 'has worksheet'
        has_workstep = 'has workstep'
        current_workstep = 'current workstep'
        template_workstep = 'template workstep'
        has_parameter = 'has parameter'
        uses_function = 'uses function'
        associated_to = 'associated to'
        # DEPRECATED, Deprecated. See SeeqNames.java for more info
        access_for = 'access for'
        # DEPRECATED, Deprecated. See SeeqNames.java for more info
        has_ace = 'has access control entry'
        has_pinned = 'has pinned'
        measuring = 'measuring'
        has_aggregation_signal = 'has aggregation signal'
        has_aggregation_condition = 'has aggregation condition'
        has_aggregation_formula_function = 'has aggregation formula function'
        bounded_by = 'bounded by'
        has_threshold = 'has threshold'
        has_condition_threshold = 'has condition threshold'
        home_folder = 'home folder'
        has_manager = 'has manager'
        package_contains_u_d_f = 'package contains user defined function'
        package_contains_doc = 'package contains formula doc'
        documented_by = 'documented by'
        agent_contains_connector = 'agent contains connector'
        connection_created_by_connector = 'connection created by connector'
        renders = 'renders'
        runs_notebook = 'runs notebook'
        source_worksheet = 'source worksheet'
        source_workstep = 'source workstep'
        has_asset = 'has asset'
        has_asset_selection = 'has asset selection'
        has_asset_selections = 'has asset selections'
        has_condition = 'has condition'
        has_item_finder = 'has item finder'
        has_content = 'has content'
        has_date_range = 'has date range'
        has_date_ranges = 'has date ranges'
        has_designated_scim_user = 'has designated scim user'
        has_scheduled_notebook = 'has scheduled notebook'
        has_template = 'has template'
        subscribed_to = 'subscribed to'
        contains_context_for = 'contains context for'
    
    class Types:
        asset = 'Asset'
        asset_selection = 'AssetSelection'
        access_control_entry = 'AccessControlEntry'
        stored_signal = 'StoredSignal'
        datasource = 'Datasource'
        datafile = 'Datafile'
        calculated_signal = 'CalculatedSignal'
        formula_item = 'FormulaItem'
        signal = 'Signal'
        tree_definition = 'TreeDefinition'
        identity = 'Identity'
        user = 'User'
        user_group = 'UserGroup'
        access_key = 'AccessKey'
        system_state = 'SystemState'
        export = 'Export'
        project_link = 'ProjectLink'
        add_on_tool = 'AddOnTool'
        base_annotation = 'BaseAnnotation'
        report = 'Report'
        report_template = 'ReportTemplate'
        journal = 'Journal'
        reply = 'Reply'
        item_finder = 'ItemFinder'
        condition = 'Condition'
        condition_monitor = 'ConditionMonitor'
        capsule = 'Capsule'
        stored_condition = 'StoredCondition'
        calculated_condition = 'CalculatedCondition'
        scalar = 'ScalarItem'
        literal_scalar = 'LiteralScalar'
        calculated_scalar = 'CalculatedScalar'
        formula_function = 'FormulaFunction'
        aggregating_formula_function = 'AggregatingFormulaFunction'
        chart = 'Chart'
        user_defined_formula_function = 'UserDefinedFormulaFunction'
        formula_package = 'FormulaPackage'
        formula_doc = 'FormulaDoc'
        formula_package_doc = 'FormulaPackageDoc'
        threshold_metric = 'ThresholdMetric'
        display = 'Display'
        display_template = 'DisplayTemplate'
        folder = 'Folder'
        project = 'Project'
        content = 'Content'
        image_content = 'ImageContent'
        react_json_content = 'ReactJsonContent'
        date_range = 'DateRange'
        scheduled_notebook = 'ScheduledNotebook'
        add_on = 'AddOn'
        plugin = 'Plugin'
        connection = 'Connection'
        connector = 'Connector'
        agent = 'Agent'
        analysis = 'Analysis'
        topic = 'Topic'
        vantage = 'Vantage'
        worksheet = 'Worksheet'
        workstep = 'Workstep'
        item = 'Item'
        item_with_owner = 'ItemWithOwner'
        type_definition = 'TypeDefinition'
        table_definition = 'TableDefinition'
        version_history = 'VersionHistory'
    
    class TopicDocumentAttributes:
        data_seeq_content = 'data-seeq-content'
        data_seeq_content_pending = 'data-seeq-content-pending'
        data_seeq_date_range_id = 'data-date-range-id'
        data_seeq_date_range_format = 'data-date-range-format'
        data_seeq_date_range_content = 'data-date-range-content'
        data_seeq_content_percent_width = 'data-seeq-content-percent-width'
        data_seeq_content_no_margin = 'data-seeq-content-no-margin'
        data_seeq_content_border = 'data-seeq-content-border'
        data_seeq_content_height = 'data-seeq-content-height'
        data_seeq_content_width = 'data-seeq-content-width'
        data_seeq_content_property_overrides = 'data-seeq-content-property-overrides'
        data_seeq_asset_selection_id = 'data-asset-selection-id'
        data_seeq_asset_selection_depth_level = 'data-asset-selection-depth-level'
        data_seeq_iframe_content_url = 'data-iframe-content-url'
        data_seeq_iframe_content_height = 'data-iframe-content-height'
        data_seeq_iframe_content_width = 'data-iframe-content-width'
        data_seeq_markdown_content = 'data-markdown-content'
    
    class CapsuleProperties:
        start = 'Start'
        end = 'End'
        duration = 'Duration'
        is_uncertain = 'IsUncertain'
        is_bounded = 'IsBounded'
        middle = 'Middle'
        count = 'Count'
        value = 'Value'
        unit = 'Unit'
        distance = 'Distance'
        similarity = 'Similarity'
        standard_deviation = 'Standard Deviation'
        sample_period = 'Sample Period'
        reference_capsule = 'Reference Capsule'
        condition_id = 'Condition ID'
        capsule_id = 'Capsule ID'
        capsule_id_safe = 'Capsule SortKey'
        original_uncertainty = 'Original Uncertainty'
        is_suppressed = 'Is Suppressed'
    
    class Properties:
        guid = 'GUID'
        name = 'Name'
        id = 'ID'
        description = 'Description'
        unsearchable = 'Unsearchable'
        archived = 'Archived'
        archival_id = 'Archival ID'
        archived_at = 'Archived At'
        archived_reason = 'Archived Reason'
        dashboard = 'Dashboard'
        sync_token = 'Sync Token'
        number_format = 'Number Format'
        source_number_format = 'Source Number Format'
        string_format = 'String Format'
        source_string_format = 'Source String Format'
        storage_location = 'Storage Location'
        executor_id = 'Executor ID'
        creator_id = 'Creator ID'
        creator_first_name = 'Creator First Name'
        creator_last_name = 'Creator Last Name'
        historical_data_version = 'Historical Data Version'
        manual_cdc = 'Manual CDC'
        scoped_to = 'Scoped To'
        asset = 'asset'
        full_path = 'fullpath'
        locked = 'Locked'
        locked_metadata = 'Locked Metadata'
        import_item_type = 'Import Item Type'
        filename = 'Filename'
        name_row = 'Name Row'
        name_prefix = 'Name Prefix'
        name_suffix = 'Name Suffix'
        condition_name = 'Condition Name'
        first_data_row = 'First Data Row'
        description_row = 'Description Row'
        interpolation_method_row = 'Interpolation Method Row'
        maximum_interpolation_row = 'Maximum Interpolation Row'
        value_uom_row = 'Value Unit of Measure Row'
        field_delimiter = 'Field Delimiter'
        day_first_default = 'Day First Default'
        key_column_name = 'Key Column Name'
        key_column_index = 'Key Column Index'
        end_column_name = 'End Column Name'
        end_column_index = 'End Column Index'
        key_format = 'Key Format'
        time_zone = 'Time Zone'
        value_column_names = 'Value Column Names'
        value_column_indices = 'Value Column Indices'
        append = 'Append'
        validation_mode = 'Validation Mode'
        lenient_daylight_savings = 'Lenient Daylight Savings'
        datasource_class = 'Datasource Class'
        datasource_name = 'Datasource Name'
        datasource_id = 'Datasource ID'
        datasource_host_id = 'Datasource Host ID'
        data_id = 'Data ID'
        previous_data_id = 'Previous Data ID'
        stored_in_seeq = 'Stored In Seeq'
        connection_request_timeout = 'Connection Request Timeout'
        max_requests = 'Max Requests'
        windows_auth_mode = 'Windows Auth Mode'
        ldap_auth_mode = 'LDAP Auth Mode'
        oidc_issuer = 'Oidc Issuer'
        oidc_jwt_key_set = 'Oidc JWT Key Set'
        oidc_audience = 'Oidc Audience'
        username_password_compatible = 'Username/Password Compatible'
        username_human_readable = 'Username Human Readable'
        expect_duplicates_during_indexing = 'Expect Duplicates During Indexing'
        referenced_datasource_id = 'Referenced Datasource ID'
        referenced_datasource_class = 'Referenced Datasource Class'
        additional_groups_to_sync = 'Additional Groups To Sync'
        signal_location = 'Signal Location'
        condition_location = 'Condition Location'
        current_index_at = 'Current Index At'
        previous_index_at = 'Previous Index At'
        sync_result = 'Sync Result'
        indexing_schedule_supported = 'Indexing Schedule Supported'
        sync_mode = 'Sync Mode'
        items_archived_count = 'Items Archived Count'
        indexing_duration = 'Indexing Duration'
        provides_everyone_group_identities = 'Provides Everyone Group Identities'
        datasource_labels = 'Datasource Labels'
        datasource_vendor = 'Datasource Vendor'
        indexing_progress_timestamp = 'Indexing Progress Timestamp'
        asset_progress = 'Asset Progress'
        asset_count = 'Asset Count'
        previous_asset_count = 'Previous Asset Count'
        signal_progress = 'Signal Progress'
        signal_count = 'Signal Count'
        previous_signal_count = 'Previous Signal Count'
        condition_progress = 'Condition Progress'
        condition_count = 'Condition Count'
        previous_condition_count = 'Previous Condition Count'
        scalar_progress = 'Scalar Progress'
        scalar_count = 'Scalar Count'
        previous_scalar_count = 'Previous Scalar Count'
        relationship_progress = 'Relationship Progress'
        relationship_count = 'Relationship Count'
        previous_relationship_count = 'Previous Relationship Count'
        user_group_progress = 'UserGroup Progress'
        user_group_count = 'UserGroups Count'
        previous_user_group_count = 'Previous UserGroups Count'
        stale_metadata = 'Stale Metadata'
        interpolation_method = 'Interpolation Method'
        source_maximum_interpolation = 'Source Maximum Interpolation'
        override_maximum_interpolation = 'Override Maximum Interpolation'
        maximum_interpolation = 'Maximum Interpolation'
        value_uom = 'Value Unit Of Measure'
        source_value_uom = 'Source Value Unit Of Measure'
        display_unit = 'Display Unit'
        units_migrated = 'Units Migrated'
        maximum_duration = 'Maximum Duration'
        cache_enabled = 'Cache Enabled'
        precaching_enabled = 'Pre-Caching Enabled'
        uncertainty_override = 'Uncertainty Override'
        metadata_properties = 'Metadata Properties'
        last_key_written = 'Last Key Written'
        request_interval_clamp = 'Request Interval Clamp'
        old_cache_id = 'Old Cache ID'
        debounce_duration = 'Debounce Duration'
        debounce_window = 'Debounce Window'
        tombstoned_properties = 'Tombstoned Properties'
        multivalued_samples_behavior = 'Multivalued Samples Behavior'
        capsule_id_property = 'Capsule ID Property'
        stored_series_cache_version = 'Stored Series Cache Version'
        replace_capsule_properties = 'Replace Capsule Properties'
        uom = 'Unit Of Measure'
        formula_version = 'Formula Version'
        formula = 'Formula'
        swap_source_id = 'Swap Source ID'
        swap_source_uom = 'Source Value Unit Of Measure'
        swap_key = 'Swap Key'
        cache_value = 'CacheValue'
        value = 'Value'
        formula_parameters = 'FormulaParameters'
        is_modified_outside_table = 'Is Modified Outside Table'
        expected_provided_type = 'ExpectedProvidedType'
        version_string = 'Version String'
        creator_name = 'Creator Name'
        creator_contact_info = 'Creator Contact Info'
        title = 'Title'
        examples = 'Examples'
        search_keywords = 'Search Keywords'
        experimental_package_name = 'Experimental'
        seeq_package_name = 'Seeq'
        index_doc_name = 'index'
        aggregation_function = 'AggregationFunction'
        duration = 'Duration'
        period = 'Period'
        measured_item_maximum_duration = 'MeasuredItemMaximumDuration'
        measured_item_id = 'MeasuredItemId'
        metric_configuration_migration = 'MetricReconfigurationMigration'
        bounding_condition_maximum_duration = 'BoundingConditionMaximumDuration'
        neutral_color = 'NeutralColor'
        is_generated = 'Is Generated'
        tree_type = 'Tree Type'
        manually_added = 'Manually Added'
        created_by = 'Created By'
        asset_path = 'assetPath'
        column_type = 'columnType'
        original_mapping = 'originalMapping'
        parameter_to_column = 'parameterToColumn'
        swap_source_asset_id = 'Swap Source Asset ID'
        export_data = 'Export Data'
        is_capsule_view = 'Is Capsule View'
        is_export_capsule_table = 'Is Export Capsule Table'
        is_chain_view = 'Is Chain View'
        is_autoupdate_time_range = 'Is Autoupdate Time Range'
        grid_size = 'Grid Size'
        user_email = 'User Email'
        username = 'Username'
        first_name = 'First Name'
        last_name = 'Last Name'
        password_hash = 'Password Hash'
        password_hash_method = 'Password Hash Method'
        password_set_date = 'Password Set Date'
        last_login_at = 'Last Login At'
        failed_login_attempt_count = 'Failed Login Attempt Count'
        session_duration = 'Session Duration'
        access_key_validity_duration = 'Access Key Validity Duration'
        enabled = 'Enabled'
        workbench = 'Workbench'
        user_role = 'User Role'
        site = 'Site'
        user_language = 'userLanguage'
        require_password_reset_on_next_login = 'Require Password Reset On Next Login'
        agent_provisioning_status = 'Agent Provisioning Status'
        agent_provisioning_source_address = 'Agent Provisioning Source Address'
        agent_provisioning_requested_at = 'Agent Provisioning Requested At'
        password_reset_token_hash = 'Password Reset Token Hash'
        password_reset_token_expiration_at = 'Password Reset Token Expiration At'
        last_password_reset_request_at = 'Last Password Reset Request At'
        remote_group_editable = 'Remote Group Editable'
        remove_permissions = 'Remove Permissions'
        data = 'Data'
        created_at = 'Created At'
        updated_at = 'Updated At'
        order = 'Order'
        removed_worksheet_workstep_relationships = 'Removed Worksheet Workstep Relationships'
        document = 'Document'
        plain_text_document = 'Plain Text Document'
        discoverable = 'Discoverable'
        published_at = 'Published At'
        last_viewed_at = 'Last Viewed At'
        source_security_string = 'Source Security String'
        security_string = 'Security String'
        permissions_from_datasource = 'Permissions From Datasource'
        permission_inheritance_disabled = 'Permission Inheritance Disabled'
        asset_group_member = 'Asset Group Member'
        first_admin_created = 'First Admin Created'
        u_i_config = 'UIConfig'
        dirty_cache = 'Dirty Cache'
        unmodifiable = 'Unmodifiable'
        height = 'Height'
        width = 'Width'
        scale = 'Scale'
        timezone = 'Timezone'
        selector = 'Selector'
        last_screenshot_duration = 'Last Screenshot Duration'
        screenshot_warning = 'Screenshot Warning'
        summary_type = 'Summary Type'
        summary_value = 'Summary Value'
        hide_uncertainty = 'Hide Uncertainty'
        current_filename = 'Current Filename'
        last_result_hash = 'Last Result Hash'
        cache_id = 'Cache ID'
        asset_path_depth = 'Asset Path Depth'
        last_run_state = 'Last Run State'
        query_range_look_ahead = 'Query Range Look Ahead'
        webhook_url = 'Webhook Url'
        first_run_look_back = 'First Run Look Back'
        capsule_event_types = 'Capsule Event Types'
        finder_configurations = 'Finder Configurations'
        include_hierarchy = 'Include Hierarchy'
        resource_size = 'Resource Size'
        project_type = 'Project Type'
        startup_environment = 'Startup Environment'
        binary_data = 'Binary Data'
        last_run_at = 'Last Run At'
        # DEPRECATED, Deprecated. See SeeqNames.java for more info
        previous_run_time = 'Previous Run Time'
        scheduled_at = 'Scheduled At'
        total_run_time = 'Total Run Time'
        average_run_time = 'Average Run Time'
        cron_schedule = 'Cron Schedule'
        background = 'Background'
        condition_formula_now = 'Condition Formula Now'
        last_run_warnings = 'Last Run Warnings'
        last_run_errors = 'Last Run Errors'
        notebook_path = 'Notebook Path'
        cron_schedules_to_indices = 'Cron Schedules To Indices'
        label = 'Label'
        notify_on_skipped_execution = 'Notify On Skipped Execution'
        notify_on_automatic_unschedule = 'Notify On Automatic Unschedule'
        is_fixed_with = 'Is Fixed Width'
        margin_top = 'Margin Top'
        margin_bottom = 'Margin Bottom'
        margin_left = 'Margin Left'
        margin_right = 'Margin Right'
        orientation = 'Page Orientation'
        page_size = 'Page Size'
        is_ck_enabled = 'Is CK Enabled'
        froala_backup = 'Froala Backup'
        original_has_workstep_guid = 'Original HasWorkstep GUID'
        in_development = 'In Development'
        icon_class = 'Icon Class'
        target_url = 'Target URL'
        launch_location = 'Launch Location'
        link_type = 'Link Type'
        window_details = 'Window Details'
        reuse_window = 'Reuse Window'
        sort_key = 'Sort Key'
        add_on_identifier = 'Add-on Identifier'
        add_on_components = 'Add-on Components'
        html_template = 'HTML Template'
        plain_text_template = 'Plain Text Template'
        category = 'Category'
        identifier = 'Identifier'
        location = 'Location'
        entry_point = 'Entry Point'
        icon = 'Icon'
        last_update_time = 'Last Update Time'
        host = 'Host'
        options = 'Options'
        version = 'Version'
        json = 'Json'
        propagation_pending = 'Propagation Pending'
        datasource_managed = 'Datasource Managed'
        connection_id = 'ConnectionId'
        indexing_connection = 'Indexing Connection'
        pull_connection = 'Pull Connection'
        max_concurrent_requests = 'Max Concurrent Requests'
        max_results_per_requests = 'Max Results Per Requests'
        transforms = 'Transforms'
        type = 'Type'
        indexing_frequency = 'Indexing Frequency'
        next_scheduled_index_at = 'Next Scheduled Index At'
        disabled_at = 'Disabled At'
        disabled_by = 'Disabled By'
        indexing_on_startup_and_config_change = 'Indexing On Startup And Config Change'
        column_definitions = 'Column Definitions'
        batch_action = 'Batch Action'
        table_type = 'Table Type'
        is_context_condition = 'Is Context Condition'
        versioned_item_type = 'Versioned Item Type'
        versions = 'Versions'
        inputs = 'Inputs'
        result_type = 'Result Type'
    






from __future__ import annotations

import math
import textwrap
from dataclasses import dataclass
from datetime import timedelta
from typing import Dict

from seeq.base.seeq_names import SeeqNames


@dataclass
class RequestTimings:
    """
    Direct port of net-link/sdk/Seeq.Link.SDK/Utilities/RequestTimings.cs
    """
    datasource_duration: timedelta
    datasource_samples_count: int
    datasource_capsules_count: int
    cache_duration: timedelta
    cache_persisted_samples_count: int
    cache_persisted_capsules_count: int
    cache_in_memory_samples_count: int
    cache_in_memory_capsules_count: int
    calc_engine_processing_duration: timedelta
    calc_engine_queue_duration: timedelta
    request_queue_duration: timedelta
    garbage_collection_duration: timedelta
    metadata_duration: timedelta
    metadata_items_count: int
    metadata_relationships_count: int

    # NOTE: since appserver always sends numeric data formatted in the US locale without thousands separators, there
    # is no need for a locale-specific parsing logic
    @staticmethod
    def from_milliseconds_string(milliseconds: str) -> timedelta:
        return timedelta(milliseconds=float(milliseconds))

    @staticmethod
    def parse_meter(header_value: str) -> int:
        return int(float(header_value))

    @staticmethod
    def from_api_response_headers(headers: Dict[str, str]) -> RequestTimings:
        pieces = dict()

        for header in [headers[SeeqNames.API.Headers.server_timing], headers[SeeqNames.API.Headers.server_meters]]:
            for _tuple in header.split(','):
                try:
                    parts = dict()
                    for part in _tuple.split(';'):
                        if part.find('=') == -1:
                            continue

                        key = part[:part.find('=')]
                        value = part[part.find('=') + 1:]
                        parts[key] = value

                    if 'dur' not in parts or 'desc' not in parts:
                        continue

                    pieces[parts['desc'].replace('"', '')] = parts['dur']
                except KeyError:
                    # CRAB-40649 There may be non-Seeq headers in the Server-Timings. Allow the rest of the pieces
                    # to complete even if this segment was invalid.
                    pass

        def _safely_get_timedelta_for_header(_key: str):
            if _key in pieces:
                try:
                    return RequestTimings.from_milliseconds_string(pieces[_key])
                except ValueError:
                    pass
            return timedelta()

        def _safely_get_meter_for_header(_key: str):
            if _key in pieces:
                try:
                    return RequestTimings.parse_meter(pieces[_key])
                except ValueError:
                    pass
            return 0

        return RequestTimings(
            _safely_get_timedelta_for_header(SeeqNames.API.Headers.Timings.datasource),
            _safely_get_meter_for_header(SeeqNames.API.Headers.Meters.datasource_samples_read),
            _safely_get_meter_for_header(SeeqNames.API.Headers.Meters.datasource_capsules_read),
            _safely_get_timedelta_for_header(SeeqNames.API.Headers.Timings.cache),
            _safely_get_meter_for_header(SeeqNames.API.Headers.Meters.cache_samples_read),
            _safely_get_meter_for_header(SeeqNames.API.Headers.Meters.cache_capsules_read),
            _safely_get_meter_for_header(SeeqNames.API.Headers.Meters.cache_in_memory_samples_read),
            _safely_get_meter_for_header(SeeqNames.API.Headers.Meters.cache_in_memory_capsules_read),
            _safely_get_timedelta_for_header(SeeqNames.API.Headers.Timings.processing),
            _safely_get_timedelta_for_header(SeeqNames.API.Headers.Timings.calc_engine_queue),
            _safely_get_timedelta_for_header(SeeqNames.API.Headers.Timings.request_queue),
            _safely_get_timedelta_for_header(SeeqNames.API.Headers.Timings.g_c),
            _safely_get_timedelta_for_header(SeeqNames.API.Headers.Timings.seeq_database),
            _safely_get_meter_for_header(SeeqNames.API.Headers.Meters.database_items_read),
            _safely_get_meter_for_header(SeeqNames.API.Headers.Meters.database_relationships_read))

    def __add__(self, other):
        return RequestTimings(
            self.datasource_duration + other.datasource_duration,
            self.datasource_samples_count + other.datasource_samples_count,
            self.datasource_capsules_count + other.datasource_capsules_count,
            self.cache_duration + other.cache_duration,
            self.cache_persisted_samples_count + other.cache_persisted_samples_count,
            self.cache_persisted_capsules_count + other.cache_persisted_capsules_count,
            self.cache_in_memory_samples_count + other.cache_in_memory_samples_count,
            self.cache_in_memory_capsules_count + other.cache_in_memory_capsules_count,
            self.calc_engine_processing_duration + other.calc_engine_processing_duration,
            self.calc_engine_queue_duration + other.calc_engine_queue_duration,
            self.request_queue_duration + other.request_queue_duration,
            self.garbage_collection_duration + other.garbage_collection_duration,
            self.metadata_duration + other.metadata_duration,
            self.metadata_items_count + other.metadata_items_count,
            self.metadata_relationships_count + other.metadata_relationships_count)


class Usage:
    """
    Used during spy.pull() to track data processing and timing information that is seen in the "rocket" in Seeq
    Workbench. This class is returned in the "Data Processed" column of the Status DataFrame of spy.pull calls.
    """

    request_timings: RequestTimings

    def __init__(self, _bytes: int = 0):
        self.request_timings = RequestTimings(
            timedelta(),
            0,
            0,
            timedelta(),
            0,
            0,
            0,
            0,
            timedelta(),
            timedelta(),
            timedelta(),
            timedelta(),
            timedelta(),
            0,
            0
        )

    def __str__(self):
        return self._humanized()

    def __repr__(self):
        return (textwrap.dedent(f"""
            Time spent in Request Queue:           {self.request_timings.request_queue_duration}
            Time spent reading Metadata:           {self.request_timings.metadata_duration}
            Time spent waiting for Datasource(s):  {self.request_timings.datasource_duration}
            Time spent reading from Seeq Cache:    {self.request_timings.cache_duration}
            Time spent in Calc Engine Queue:       {self.request_timings.calc_engine_queue_duration}
            Time spent in Calc Engine:             {self.request_timings.calc_engine_processing_duration}
            Time spent reclaiming Memory:          {self.request_timings.garbage_collection_duration}

            Metadata items read:                {self.request_timings.metadata_items_count:>17}
            Metadata relationships read:        {self.request_timings.metadata_relationships_count:>17}
            Samples read from Datasource(s):    {self.request_timings.datasource_samples_count:>17}
            Samples read from Persistent Cache: {self.request_timings.cache_persisted_samples_count:>17}
            Samples read from In-Memory Cache:  {self.request_timings.cache_in_memory_samples_count:>17}
            Capsules read from Datasource(s):   {self.request_timings.datasource_capsules_count:>17}
            Capsules read from Persistent Cache:{self.request_timings.cache_persisted_capsules_count:>17}
            Capsules read from In-Memory Cache: {self.request_timings.cache_in_memory_capsules_count:>17}
            Total bytes processed:              {self.bytes_processed:>17}
        """).strip())

    @property
    def bytes_processed(self) -> int:
        # The multiplier matches
        # appserver/drivers/graph/src/main/java/com/seeq/appserver/driver/graph/service/DataConsumptionReportService.kt
        return (16 * (self.request_timings.cache_in_memory_samples_count +
                      self.request_timings.cache_persisted_samples_count +
                      self.request_timings.datasource_samples_count)

                + 64 * (self.request_timings.cache_in_memory_capsules_count +
                        self.request_timings.cache_persisted_capsules_count +
                        self.request_timings.datasource_capsules_count))

    def _humanized(self) -> str:
        if self.bytes_processed == 0:
            return '0 B'

        suffixes = ['B', 'KB', 'MB', 'GB', 'TB', 'PB']
        i = int(math.floor(math.log(self.bytes_processed, 1000)))
        p = math.pow(1000, i)
        s = int(self.bytes_processed / p)
        return '%s %s' % (s, suffixes[i])

    def add(self, http_headers: Dict[str, str]):
        request_timings = RequestTimings.from_api_response_headers(http_headers)
        self.request_timings += request_timings



import os
import types
from typing import List, Optional

import numpy as np
import pandas as pd

from seeq.sdk.api import ItemsApi, TreesApi
from seeq.sdk.models import SwapInputV1, SwapOptionV1, ItemWithSwapPairsV1, ItemDependencyOutputV1, \
    ItemParameterOfOutputV1, ItemPreviewWithAssetsV1
from seeq.spy import _common, _login
from seeq.spy._errors import *
from seeq.spy._session import Session
from seeq.spy._status import Status


@Status.handle_keyboard_interrupt()
def swap(items: pd.DataFrame, assets: pd.DataFrame, *, partial_swaps_ok: bool = False,
         old_asset_format: Optional[bool] = None, errors: Optional[str] = None, quiet: Optional[bool] = None,
         status: Optional[Status] = None, session: Optional[Session] = None) -> pd.DataFrame:
    """
    Operates on a DataFrame of items by swapping out the assets that those
    items are based on. The returned DataFrame can be supplied to
    spy.pull() to retrieve the resulting data.

    Parameters
    ----------
    items : {pd.DataFrame}
        A DataFrame of items over which to perform the swapping operation. The
        only required column is ID. Typically, you will have derived this
        DataFrame via a spy.search() or spy.push(metadata) call.

    assets : {pd.DataFrame}
        A DataFrame of Asset items (and ONLY Asset items) to swap IN. Each row
        must have valid ID, Type, Path, Asset, and Name columns. Typically, you
        will have derived this DataFrame via a spy.search() call.

        When a calculation depends on multiple assets, you must specify Swap
        Groups (with the notable exception of a "multi-level swap", see below.
        You can determine the set of "swappable assets" by invoking
        spy.search(include_swappable_assets=True). Then you must assemble an
        assets DataFrame where each row is the asset to be swapped in and there
        is a "Swap Out" column that specifies the corresponding asset (from
        the list of swappable assets) to be swapped out. The "Swap Out" column
        can be a string ID, the latter portion of an asset path/name, or a
        DataFrame row representing the asset (which could be used directly
        from the "Swappable Assets" column of your search result).
        Additionally, assuming you want to produce several swapped
        calculations, you must group the swap rows together by specifying a
        "Swap Group" column where unique values are used to group together
        the rows that comprise the set of assets to be swapped in/out.

        If a calculation depends on asset(s) where the immediate
        parent asset is a "categorical" name (e.g. "Raw" and "Cleansed"), it
        is referred to as a "multi-level swap", where you actually wish to
        swap at a level higher than the immediate parent. You can do so by
        specifying that higher level, and spy will automatically figure out
        the actual lower-level items that are appropriate to swap.

    partial_swaps_ok : bool, default False
        If True, allows partial swaps to occur. A partial swap occurs when
        the incoming asset has children that only partially match the outgoing
        asset.

    old_asset_format : bool, default True
        If your DataFrame doesn't use the "old" asset format, you can specify
        False for this argument. See spy.search() documentation for more info.

    quiet : bool, default False
        If True, suppresses progress output. Note that when status is
        provided, the quiet setting of the Status object that is passed
        in takes precedence.

    errors : {'raise', 'catalog'}, default 'raise'
        If 'raise', any errors encountered will cause an exception. If 'catalog',
        errors will be added to a 'Result' column in the status.df DataFrame
        (or 'Swap Result' column if using `spy.options.compatibility = 189` or lower).

    status : spy.Status, optional
        If specified, the supplied Status object will be updated as the command
        progresses. It gets filled in with the same information you would see
        in Jupyter in the blue/green/red table below your code while the
        command is executed. The table itself is accessible as a DataFrame via
        the status.df property.

    session : spy.Session, optional
        If supplied, the Session object (and its Options) will be used to
        store the login session state. This is useful to log in to different
        Seeq servers at the same time or with different credentials.

    Returns
    -------
    pandas.DataFrame
        A DataFrame with rows for each item swapped. Includes a "Result"
        column that is either "Success" or an error message (when
        errors='catalog'). Also includes a "Swap Performed" column that
        specifies the details of what swap pairs were utilized.

        Additionally, the following properties are stored on the "spy"
        attribute of the output DataFrame:

        =================== ===================================================
        Property            Description
        =================== ===================================================
        func                A str value of 'spy.swap'
        kwargs              A dict with the values of the input parameters
                            passed to spy.swap to get the output DataFrame
        status              A spy.Status object with the status of the
                            spy.swap call
        =================== ===================================================

    """
    input_args = _common.validate_argument_types([
        (items, 'items', pd.DataFrame),
        (assets, 'assets', pd.DataFrame),
        (partial_swaps_ok, 'partial_swaps_ok', bool),
        (old_asset_format, 'old_asset_format', bool),
        (errors, 'errors', str),
        (quiet, 'quiet', bool),
        (status, 'status', Status),
        (session, 'session', Session)
    ])

    session = Session.validate(session)
    status = Status.validate(status, session, quiet, errors)
    _login.validate_login(session, status)

    old_asset_format = _common.resolve_old_asset_format_arg(old_asset_format, assets)
    # Swap originally used 'Swap Result' as the status output, but 'Result' is more consistent with the rest of SPy.
    result_key = 'Swap Result' if session.options.wants_compatibility_with(189) else 'Result'

    items_api = ItemsApi(session.client)

    if 'ID' not in items.columns:
        raise ValueError('"ID" column not found in items DataFrame')

    for col in ['ID', 'Type', 'Path', 'Asset', 'Name']:
        if col not in assets.columns:
            raise ValueError(f'"{col}" column not found in assets DataFrame')

    status.update(f'Performing swaps for {len(items)} items across {len(assets)} assets', Status.RUNNING)

    item_list = list()
    for index, row in items.iterrows():
        # noinspection PyBroadException
        try:
            if not _common.present(row, 'ID'):
                raise Exception(f'items DataFrame contains row without valid ID at index {index}')

            item_dict = dict(row.to_dict())

            item_id = _common.get(row, 'ID')
            item_dependency_output = items_api.get_formula_dependencies(id=item_id)
            swappable_assets = get_swappable_assets(item_dependency_output)

            if len(swappable_assets) == 0:
                item_dict[result_key] = ("This item does not depend on any assets, either as the item's parent or "
                                         "as the parent of its formula parameters.")
                continue

            item_list.append((item_dependency_output, swappable_assets))

        except Exception:
            if status.errors == 'raise':
                raise
            else:
                pass

    if 'Swap Group' not in assets:
        assets['Swap Group'] = np.arange(len(assets))

    swap_groups = list(dict.fromkeys(assets['Swap Group'].to_list()))
    row_index = 0
    staged_swaps = [pd.DataFrame({'Original ID': pd.Series(dtype='str'),
                                  'ID': pd.Series(dtype='str'),
                                  'Type': pd.Series(dtype='str'),
                                  'Name': pd.Series(dtype='str'),
                                  result_key: pd.Series(dtype='str'),
                                  'Swap Performed': pd.Series(dtype='str')})]
    for swap_group in swap_groups:
        swap_rows = assets[assets['Swap Group'] == swap_group]

        for item_dependency_output, dependencies_with_relevant_assets in item_list:
            staged_swaps.append(pd.DataFrame({
                'Original ID': item_dependency_output.id,
                'ID': np.nan,
                'Type': item_dependency_output.type,
                'Name': item_dependency_output.name,
                result_key: 'Pending',
                'Swap Performed': np.nan
            }, index=[0]))

            def _on_success(_row_index, _job_result):
                for _key, _value in _job_result.items():
                    status.df.at[_row_index, _key] = _value

            status.add_job(row_index, (_do_swap, session, dependencies_with_relevant_assets,
                                       item_dependency_output,
                                       old_asset_format,
                                       partial_swaps_ok, swap_rows), _on_success)

            row_index += 1

    status.df = pd.concat(staged_swaps, ignore_index=True)
    status.execute_jobs(session)

    _construct_header_column(status.df)

    push_df_properties = types.SimpleNamespace(
        func='spy.swap',
        kwargs=input_args,
        status=status)

    result_df: pd.DataFrame = status.df.copy()

    result_df.drop_duplicates(subset=['Original ID', 'ID'], inplace=True)

    _common.put_properties_on_df(result_df, push_df_properties)

    error_count = len(result_df[result_df[result_key] != 'Success'])
    success_count = len(result_df[result_df[result_key] == 'Success'])
    if error_count > 0:
        status.update(f'Returned {success_count} swapped item(s). {error_count} item(s) had '
                      f'errors, see "Result" column.', Status.FAILURE)
    else:
        status.update(f'Success: Returned {success_count} swapped item(s).', Status.SUCCESS)

    return result_df


def _do_swap(session: Session, dependencies_with_relevant_assets, item_dependency_output, old_asset_format,
             partial_swaps_ok, swap_rows):
    items_api = ItemsApi(session.client)
    result_key = 'Swap Result' if session.options.wants_compatibility_with(189) else 'Result'

    swap_result = {
        'Original ID': item_dependency_output.id,
        'ID': np.nan,
        'Type': item_dependency_output.type,
        'Name': item_dependency_output.name,
        result_key: 'Success'
    }

    swap_list: List[SwapInputV1] = list()
    for index, swap_row in swap_rows.iterrows():
        if _common.get(swap_row, 'Type', 'Asset') != 'Asset':
            raise SPyValueError(f'assets DataFrame contains non-Asset type at row index {index}')

        swap_out = _common.get(swap_row, 'Swap Out')
        if swap_out is None:
            if len(dependencies_with_relevant_assets) == 1:
                swap_out = dependencies_with_relevant_assets[0].ancestors[-1].id
            else:
                raise SPyRuntimeError(
                    f'Item "{item_dependency_output.name}" ({item_dependency_output.id}) depends on '
                    'multiple assets. You must supply "Swap Group" and "Swap Out" columns in the '
                    'DataFrame, see spy.swap() function documentation for details.')

        try:
            swap_inputs = _create_swap_inputs(session, item_dependency_output.id, swap_row,
                                              old_asset_format, swap_out,
                                              dependencies_with_relevant_assets,
                                              partial_swaps_ok)
        except Exception as e:
            raise SPyRuntimeError(f'Error attempting to create swap for item '
                                  f'"{item_dependency_output.name}" ({item_dependency_output.id}):\n'
                                  f'{_common.format_exception(e)}')

        swap_list.extend(swap_inputs)

    try:
        swapped_item = items_api.find_swap(id=item_dependency_output.id, body=swap_list)
    except ApiException as e:
        raise SPyRuntimeError(f'Error finding swap for item '
                              f'"{item_dependency_output.name}" ({item_dependency_output.id}):\n'
                              f'{_common.format_exception(e)}')

    swap_result['ID'] = swapped_item.id
    swap_result['Swap Performed'] = '\n'.join([
        (getattr(s, 'friendly_swap_out_name') + ' --> ' + getattr(s, 'friendly_swap_in_name'))
        for s in swap_list
    ])

    return swap_result


def _construct_header_column(df: pd.DataFrame):
    verbose_headers = list()
    results = df.to_dict('records')
    for result in results:
        if pd.isna(result['Swap Performed']):
            continue

        for swap_instance in result['Swap Performed'].split('\n'):
            to_from = swap_instance.split(' --> ')
            verbose_headers.append(to_from[0] + ' >> ' + result['Name'])
            verbose_headers.append(to_from[1] + ' >> ' + result['Name'])
    new_headers = remove_common_prefix_suffix(verbose_headers)
    new_header_map = dict(zip(verbose_headers, new_headers))
    header_column = list()
    for result in results:
        if pd.isna(result['Swap Performed']):
            header_column.append(np.nan)
            continue

        new_name_pieces = list()
        for swap_instance in result['Swap Performed'].split('\n'):
            path_list = swap_instance.split(' --> ')[-1] + ' >> ' + result['Name']
            new_name_pieces.append(new_header_map[path_list])
        header_column.append(', '.join(new_name_pieces))

    df['Header'] = pd.Series(header_column, dtype=object)


def _create_swap_inputs(session: Session, item_id, swap_in_row, old_asset_format, swap_out,
                        possibilities, partial_swaps_ok) -> List[SwapInputV1]:
    swap_in_id = swap_in_row['ID']

    matched_possibilities = list()
    for possibility in possibilities:
        if len(possibility.ancestors) == 0:
            continue

        if _common.is_guid(swap_out):
            if swap_out == possibility.ancestors[-1].id:
                matched_possibilities.append(possibility)
                break

            continue

        elif isinstance(swap_out, str):
            swap_out_path_list = _common.path_string_to_list(swap_out)
            if len(swap_out_path_list) > len(possibility.ancestors):
                continue
            relevant_ancestors = [a.name for a in possibility.ancestors[-len(swap_out_path_list):]]
            if relevant_ancestors == swap_out_path_list:
                matched_possibilities.append(possibility)
                continue

        else:
            if not _common.present(swap_out, 'ID'):
                raise SPyValueError(f'"Swap Out" value does not have valid "ID":\n{swap_out}')

            if _common.get(swap_out, 'ID') == possibility.ancestors[-1].id:
                matched_possibilities.append(possibility)
                break

    if len(matched_possibilities) == 0:
        # Try to perform a multi-level swap
        items_api = ItemsApi(session.client)
        swap_options = items_api.get_swap_options(id=swap_in_id, swap_out_item_ids=[item_id])
        if len(swap_options.swap_options) == 1:
            swap_option: SwapOptionV1 = swap_options.swap_options[0]
            swap_root_candidate = swap_option.swap_root_candidate
            it_matches = False
            if _common.is_guid(swap_out):
                if swap_out == swap_root_candidate.id:
                    it_matches = True

            elif isinstance(swap_out, str):
                swap_out_path_list = _common.path_string_to_list(swap_out)
                if len(swap_out_path_list) <= (len(swap_root_candidate.ancestors) + 1):
                    relevant_ancestor_count = len(swap_out_path_list) - 1
                    relevant_ancestors = list()
                    if relevant_ancestor_count > 0:
                        relevant_ancestors += [a.name for a in swap_root_candidate.ancestors[-relevant_ancestor_count:]]
                    relevant_ancestors += [swap_root_candidate.name]
                    if relevant_ancestors == swap_out_path_list:
                        it_matches = True

            else:
                # Assume it's a dict or a Pandas series (i.e., a row)
                if not _common.present(swap_out, 'ID'):
                    raise SPyValueError(f'"Swap Out" value does not have valid "ID":\n{swap_out}')

                if _common.get(swap_out, 'ID') == swap_root_candidate.id:
                    it_matches = True

            if it_matches:
                # We take the best match of the bunch
                sorted_items_with_swap_pairs = swap_option.items_with_swap_pairs.copy()
                sorted_items_with_swap_pairs.sort(key=lambda item: item.parameter_match, reverse=True)
                chosen_pair: ItemWithSwapPairsV1 = sorted_items_with_swap_pairs[0]
                if chosen_pair.parameter_match < 1.0 and not partial_swaps_ok:
                    invalid_swap_outs_str = '\n'.join([
                        f'{invalid_swap_out.item.type} "{invalid_swap_out.item.name}" ({invalid_swap_out.item.id})'
                        for invalid_swap_out in swap_option.invalid_swap_outs
                    ])
                    raise SPyValueError(f'For swap-in row...\n\n{swap_in_row}\n\n...the best match is only '
                                        f'{int(chosen_pair.parameter_match * 100)}% similar.\n\n'
                                        f'Items we could not find swaps for:\n{invalid_swap_outs_str}\n\n'
                                        f'Specify partial_swaps_ok=True to swap in whatever matches and leave '
                                        f'the rest un-swapped.')

                swap_inputs = chosen_pair.swap_pairs
                for swap_input in swap_inputs:  # type: SwapInputV1
                    # We only have IDs, so we have to make an API call to get the ancestors
                    trees_api = TreesApi(session.client)
                    swap_out_tree_item: ItemPreviewWithAssetsV1 = trees_api.get_tree(id=swap_input.swap_out).item
                    swap_in_tree_item: ItemPreviewWithAssetsV1 = trees_api.get_tree(id=swap_input.swap_in).item

                    swap_out_path = (' >> '.join([a.name for a in swap_out_tree_item.ancestors]) +
                                     f' >> {swap_out_tree_item.name}')
                    swap_in_path = (' >> '.join([a.name for a in swap_in_tree_item.ancestors]) +
                                    f' >> {swap_in_tree_item.name}')

                    setattr(swap_input, 'friendly_swap_out_name', swap_out_path)
                    setattr(swap_input, 'friendly_swap_in_name', swap_in_path)

                # Sort the outputs by swap_out so that headers are assembled deterministically
                return sorted(swap_inputs, key=lambda s: s.swap_out)

    if len(matched_possibilities) == 0:
        swappable_assets_str = "\n".join(
            [_common.path_list_to_string([a.name for a in m.ancestors]) for m in possibilities])
        raise SPyValueError(f'"Swap Out" value...\n{swap_out}\n'
                            '...could not be matched against any of the following swappable assets:\n'
                            f'{swappable_assets_str}')

    if len(matched_possibilities) > 1:
        matched_assets_str = "\n".join(
            [_common.path_list_to_string([a.name for a in m.ancestors]) for m in matched_possibilities])
        raise SPyValueError(f'"Swap Out" value...\n{swap_out}\n...matched multiple swappable assets:\n'
                            f'{matched_assets_str}')

    chosen_possibility = matched_possibilities[0]

    swap_input = SwapInputV1(swap_in=swap_in_id, swap_out=chosen_possibility.ancestors[-1].id)
    setattr(swap_input, 'friendly_swap_out_name', ' >> '.join([a.name for a in chosen_possibility.ancestors]))
    if old_asset_format:
        setattr(swap_input, 'friendly_swap_in_name',
                f"{swap_in_row['Path']} >> {swap_in_row['Asset']} >> {swap_in_row['Name']}")
    else:
        setattr(swap_input, 'friendly_swap_in_name', f"{swap_in_row['Path']} >> {swap_in_row['Name']}")

    return [swap_input]


def get_swappable_assets(item_dependency_output: ItemDependencyOutputV1):
    items = _get_dependencies_with_relevant_assets(item_dependency_output, item_dependency_output.dependencies)
    results = list()
    ids = set()
    for item in items:
        if len(item.ancestors) == 0:
            continue

        asset = item.ancestors[-1]
        if asset.id in ids:
            continue

        ids.add(asset.id)
        results.append(item)

    return results


def _get_dependencies_with_relevant_assets(item: ItemDependencyOutputV1, dependencies: List[ItemParameterOfOutputV1]):
    # ******************************************************************************************************************
    # This is a DIRECT port of formula.utilities.ts#getDependenciesWithRelevantAssets(). Please keep this in sync with
    # that file if necessary and don't add any extra functionality to it on the Python side.
    # ******************************************************************************************************************

    # Keep track of every ID we find that matters.
    # This should end up being items with assets, or assetless leaf nodes.
    result_id_set = list()

    # TypeScript code from client/packages/webserver/app/src/utilities/formula.utilities.ts:
    #
    # find_parameters = _.chain(dependencies)
    #     .flatMap((dependency) => _.map(dependency.parameterOf, (item) => [item.id, dependency]))
    #     .groupBy(_.first)
    #     .map((pairs, dependee) => [dependee, _.map(pairs, (pair) => pair[1])])
    #     .fromPairs()
    #     .value()
    find_parameters = dict()
    for dependency in dependencies:
        for parameter_of in dependency.parameter_of:  # type: ItemPreviewWithAssetsV1
            if parameter_of.id not in find_parameters:
                find_parameters[parameter_of.id] = list()
            find_parameters[parameter_of.id].append(dependency)

    # Items whose assets we're currently trying to find out. Start with the root item.
    queue = [item]

    while len(queue) > 0:
        # take the first param.
        current = queue.pop()

        if current.id in result_id_set:
            continue

        if len(current.ancestors) > 0:
            # if it has an ancestor, put its id in the resultIds, we're done with it.
            result_id_set.append(current.id)
        else:
            # if not, add its parameters to the queue of possible ancestors
            parameters = find_parameters.get(current.id)

            if parameters is None or len(parameters) == 0:
                # leaf node; add to results so we don't bother with it anymore
                result_id_set.append(current.id)
            else:
                # not already examined, no asset ancestor - recurse to its parameters
                queue += parameters

    # Include the original item in the possible results we pick from
    # noinspection PyTypeChecker
    possible_set = dependencies + [item]

    return list(filter(lambda d: (d.id in result_id_set), possible_set))


def remove_common_prefix_suffix(fqns):
    list_of_lists = [_common.path_string_to_list(fqn) for fqn in fqns]

    common_prefix = os.path.commonprefix(list_of_lists)
    common_suffix = os.path.commonprefix([s[::-1] for s in list_of_lists])[::-1]

    common_prefix_len = len(common_prefix)
    common_suffix_len = len(common_suffix)

    if common_prefix == common_suffix:
        # Cover the edge case where the item is being swapped with itself. In such a case, remove all but
        # the last element (as part of the prefix removal) and don't do anything for suffix removal.
        common_prefix_len -= 1
        common_suffix = ''

    # Remove common prefix and suffix from each sublist
    result = [
        _common.path_list_to_string(
            inner_list[common_prefix_len: -common_suffix_len] if common_suffix else inner_list[common_prefix_len:]
        )
        for inner_list in list_of_lists
    ]

    return result



from __future__ import annotations

import glob
import os
import pathlib
import re
import shutil
import sys
import tempfile
import time
import warnings
from inspect import stack
from typing import Union


def remove_regex_capture_group_names(regex):
    # PostgreSQL will choke on named capture groups
    return re.sub(r'\?<\w+>', '', regex)


def pythonize_regex_capture_group_names(regex):
    # Unlike standard regex syntax, Python capture groups have a capital P:  (?P<group_name>.*?)
    return re.sub(r'\?<(\w+)>', r'?P<\1>', regex)


def replace_tokens_in_regex(regex, group_dict, escape=True):
    for name, value in group_dict.items():
        regex = regex.replace('${%s}' % name, re.escape(value) if escape else value)

    return regex


def os_lock(name, target, args=(), kwargs=None, timeout=7.0, retry_period=0.1):
    """
    Adapted from https://github.com/ses4j/backup/blob/master/oslockedaction.py
    To use, just wrap your python call with an os_lock() call.  Instead of:
    >>> time.sleep(5.0)
    Do:
    >>> os_lock(name="mylock", target=time.sleep, args=(5.0,))
    Nobody else using the same lockfile will be able to run at the same time.
    If timeout is None, it will never timeout if the lockfile exists.
    Otherwise it will just try to remove the lockfile after "timeout" seconds and do 'action()' anyway.
    retry_period is a float in seconds to wait between retries.
    """
    lockfile = os.path.abspath(os.path.join(tempfile.tempdir, f'{name}.lock'))

    f = None
    started_at = time.time()
    kwargs = dict() if kwargs is None else kwargs

    try:
        while True:
            if timeout is not None and time.time() > started_at + timeout:
                # timed out, it must be stale... (i hope)  so remove it.

                if time.time() > (started_at + (timeout * 2)):
                    raise RuntimeError(f"FAIL: os_lock() can't acquire directory lock, because {lockfile} was not "
                                       "cleaned up. Delete it manually if the process is finished.")

                try:
                    os.remove(lockfile)
                except OSError:
                    time.sleep(retry_period)
                    continue

            try:
                f = os.open(lockfile, os.O_RDWR | os.O_CREAT | os.O_EXCL)
                os.write(f, bytes(f'{str(__file__)} made this, called from {sys.argv[0]}', encoding='utf-8'))
                break
            except OSError:
                time.sleep(retry_period)

        target(*args, **kwargs)
    finally:
        if f:
            os.close(f)
            os.remove(lockfile)


def deprecation_warning(message, *, stack_index=2):
    frame_info = stack()
    if len(frame_info) > stack_index:
        warnings.warn(
            f'{message} (from line {frame_info[stack_index].lineno} in file "{frame_info[stack_index].filename}")')
    else:
        warnings.warn(message)


def get_platform():
    import platform
    plat = platform.platform()
    if plat.lower().find('windows') != -1:
        return 'windows'
    if plat.lower().find('linux') != -1:
        return 'linux'
    if plat.lower().find('darwin') != -1 or plat.lower().find('macos') != -1:
        return 'osx'


def is_windows():
    return get_platform() == 'windows'


def is_linux():
    return get_platform() == 'linux'


def is_osx():
    return get_platform() == 'osx'


def is_osx_aarch64():
    import platform
    return get_platform() == 'osx' and platform.platform().lower().find('arm64') != -1


def get_home_dir(user=None):
    if is_windows():
        home = os.environ['USERPROFILE']
    else:
        user_to_expand = user if user else ''
        home = os.path.expanduser(f'~{user_to_expand}')

    return home


def get_test_with_file():
    return os.path.normpath(
        os.path.join(os.path.dirname(__file__), '..', 'sdk', 'test-with.txt'))


def get_test_with_root_dir():
    test_with_file = get_test_with_file()
    test_with = os.path.normpath(os.path.join(os.path.dirname(__file__), '..', '..', '..', '..'))
    if safe_exists(test_with_file):
        with safe_open(test_with_file, 'r') as f:
            lines = f.readlines()
            test_with = lines[0]

    return test_with


def do_with_retry(func, timeout_sec=5, retry_sleep_sec=0.1):
    import time
    timer = time.time()
    while True:
        # noinspection PyBroadException
        try:
            func()
            break
        except Exception:
            if time.time() - timer > timeout_sec:
                raise

            time.sleep(retry_sleep_sec)


def safe_remove(path, *args, **kwargs):
    return os.remove(handle_long_filenames(path), *args, **kwargs)


def safe_rmtree(path, *args, **kwargs):
    return shutil.rmtree(handle_long_filenames(path), *args, **kwargs)


def safe_open(filename, *args, **kwargs):
    return open(handle_long_filenames(filename), *args, **kwargs)


def safe_makedirs(filename, *args, **kwargs):
    filename = handle_long_filenames(filename)
    # Empirically, it looks like os.makedirs() cannot handle long filenames, so throw a half-decent exception
    if len(filename) > 260:
        raise RuntimeError(f'Filename "{filename}" is too long.  Windows has a 260 character limit.')

    return os.makedirs(handle_long_filenames(filename), *args, **kwargs)


def safe_walk(filename, *args, **kwargs):
    return os.walk(handle_long_filenames(filename), *args, **kwargs)


def safe_copy(src, dst, *args, **kwargs):
    return shutil.copy(handle_long_filenames(src), handle_long_filenames(dst), *args, **kwargs)


def safe_copytree(src, dst, *args, **kwargs):
    return shutil.copytree(handle_long_filenames(src), handle_long_filenames(dst), *args, **kwargs)


def safe_isfile(filename):
    return os.path.isfile(handle_long_filenames(filename))


def safe_isdir(filename):
    return os.path.isdir(handle_long_filenames(filename))


def safe_abspath(filename):
    return os.path.abspath(handle_long_filenames(filename))


def safe_relpath(path, start):
    return os.path.relpath(handle_long_filenames(path), handle_long_filenames(start))


def safe_exists(filename):
    return os.path.exists(handle_long_filenames(filename))


def safe_glob(filename, *args, **kwargs):
    return glob.glob(handle_long_filenames(filename), *args, **kwargs)


def handle_long_filenames(path: Union[str, pathlib.Path]):
    # Make sure it is a string (could have been a pathlib.Path)
    path = str(path)

    if not is_windows():
        return path

    # Make sure the path is absolute so we can add the \\?\ prefix
    path = os.path.abspath(path)

    # Get rid of extra/unnecessary /../ or /./ sections of the path. Note that normpath does not work properly with
    # long filenames, so we have to convert to short paths and then call this before cleaning.
    path = os.path.normpath(re.sub(r'^([/\\][/\\]\?[/\\])?(\w:)', r'\2', path))

    # First replace all forward slashes with backslashes like Windows wants.
    path = re.sub(r'/', r'\\', path)

    # Windows 'long paths' are of the form \\?\D:\Foo\Bar (instead of the
    # regular D:\Foo\Bar). So we have to do a RegEx replacement that involves
    # A LOT of backslashes.
    path = re.sub(r'^(\w):\\', r'\\\\?\\\1:\\', path)

    return path


def cleanse_filename(filename, replacement_char='_'):
    return re.sub(r'[:"%\\/<>^|?*&\[\]]', replacement_char, filename)



from __future__ import annotations

"""
This module facilitates reading and writing tables of data to the console and to CSV files.
"""

import csv
import os
import time

from seeq.base import util


def start_table(
    column_specs, csv_file_name=None, append_csv=False, console=True, encoding="utf-8"
):
    table = dict()

    table["column_specs"] = column_specs
    table["csv_file_name"] = csv_file_name
    table["console"] = console

    table["csv_file"] = None
    count = 1
    try:
        table["csv_writer"] = None
        file_exists = False
        if csv_file_name:
            file_exists = util.safe_exists(csv_file_name)
            if file_exists and append_csv:
                header, rows = read_table(csv_file_name)
                count = int(rows[-1][0]) + 1

            open_command = "a+" if append_csv else "w+"
            table["csv_file"] = util.safe_open(
                csv_file_name, open_command, newline="", encoding=encoding
            )
            table["csv_writer"] = csv.writer(table["csv_file"])

        if table["console"]:
            _print_console_header(column_specs)
            print_console_underline(column_specs)

        if table["csv_writer"] and (not file_exists or not append_csv):
            _print_csv_header(table["csv_writer"], column_specs)

    except Exception:
        if table["csv_file"]:
            table["csv_file"].close()

        raise

    table["extra_values"] = {"count": count, "start_time_sec": time.time()}

    return table


def write_table_row(table, item):
    if table["console"]:
        _print_console_row(table["column_specs"], item, table["extra_values"])

    if table["csv_writer"]:
        _print_csv_row(
            table["csv_writer"], table["column_specs"], item, table["extra_values"]
        )

    table["extra_values"]["count"] = table["extra_values"]["count"] + 1


def finish_table(table):
    if table["csv_file"]:
        table["csv_file"].close()

        if table["console"]:
            print('\nOutput also written to "%s"' % table["csv_file_name"])

            hidden_columns = [
                cs.header for cs in table["column_specs"] if not cs.console
            ]
            if len(hidden_columns) > 0:
                print("with more columns: " + ", ".join(hidden_columns))


def read_table(csv_file_name):
    header = None
    rows = []

    with util.safe_open(csv_file_name, "r", encoding="utf-8-sig") as csv_file:
        # Even though (by default) we don't write tables with 'utf-8-sig' encoding, reading with 'utf-8-sig' allows to
        # handle both 'utf-8' and 'utf-8-sig'. (utf-8-sig includes a byte-order-marker at the beginning of the file.)
        csv_reader = csv.reader(csv_file)
        for row in csv_reader:
            if not header:
                header = row
            else:
                rows.append(row)

    return header, rows


def get_table_rows(table):
    header, rows = table
    return rows


def get_table_column(table, row, column_name):
    header, rows = table
    column_index = find_column_index(header, column_name)
    if column_index is None:
        raise Exception('Column "%s" not found in table' % column_name)

    return row[column_index]


def find_column_index(header, column_name):
    for i in range(0, len(header)):
        if header[i].strip() == column_name:
            return i

    return None


def print_table_from_sdk_output(
    func,
    collection_name,
    column_specs,
    action_func=None,
    csv_file_name=None,
    append_csv=False,
):
    """
    This helper method prints a nicely formatted table to the console and (optionally) outputs values to a CSV file
    by executing a Seeq SDK function that follows typical pagination with offset/limit parameters and outputs. It
    paginates through the data by executing the supplied 'func' function with a new offset.

    :param func: A function that takes a single parameter 'offset' which specifies the current offset in pagination
    :param collection_name: Paginated output always has a top-level collection field, specify it here
    :param column_specs: An array of ColumnSpec objects that define each column in the output.
    :param action_func: An optional "action function" to execute on every item in the collection
    :param csv_file_name: Optional path for a CSV file to write the output to
    :param append_csv: True to append to existing CSV file
    """
    table = None
    try:
        table = start_table(column_specs, csv_file_name, append_csv)

        iterate_over_output(
            func,
            collection_name,
            lambda item: _print_table_from_sdk_output_action(table, item, action_func),
        )

    finally:
        if table:
            finish_table(table)


class ItemSkippedException(Exception):
    def __init__(self):
        pass


def iterate_over_output(output_func, collection_name, action_func):
    """
    Executes the output_func and then iterates over the items in collection_name, making additional paginated calls
    as necessary to get through all the items.
    :param output_func: A function that takes a single parameter 'offset' which specifies the current offset in
    pagination
    :param collection_name: Paginated output always has a top-level collection field, specify it here
    :param action_func: An optional "action function" to execute on every item in the collection
    """
    offset = 0
    while True:
        try:
            output = output_func(offset)

            collection = getattr(output, collection_name)

            for item in collection:
                try:
                    action_func(item)
                except ItemSkippedException:
                    pass

            if len(collection) != output.limit:
                break

            offset += output.limit
        except Exception as e:
            raise e


class ColumnSpec:
    """
    This class contains a specification of a column in the output for print_table_from_sdk_output(). See constructor
    docs for info on what each field means.
    """

    field = ""  # type: str
    header = ""  # type: str
    align = ""  # type: str
    width = 0  # type: int
    transform = None
    console = True

    def __init__(
        self, field, header, align="<", width=16, transform=None, console=True
    ):
        """
        :param field: The field name within the returned item that will populate this column
        :type field: str
        :param header: The column header to be output in the console and in the CSV header row
        :type header: str
        :param align: Alignment for the column. '<' is left-align, '>' is right-align
        :type align: str
        :param width: The width of the column for the console output.
        :type width: int
        :param transform: A function to call to transform the field value into something else before printing.
        :type transform: function
        """
        self.field = field
        self.header = header
        self.align = align
        self.width = width
        self.transform = transform
        self.console = console


def _print_time_string(seconds):
    return "{:02}:{:04.1f}".format(int(seconds / 60 % 60), seconds % 60)


def _print_table_from_sdk_output_action(table, item, action_func=None):
    if action_func:
        start = time.time()
        action_return = action_func(item)
        table["extra_values"]["action_time_sec"] = time.time() - start
        table["extra_values"]["action_time_str"] = _print_time_string(
            table["extra_values"]["action_time_sec"]
        )
        if action_return:
            for pairs in action_return:
                extra_key, extra_value = pairs
                table["extra_values"][extra_key] = extra_value

    table["extra_values"]["total_time_sec"] = (
        time.time() - table["extra_values"]["start_time_sec"]
    )
    table["extra_values"]["total_time_str"] = _print_time_string(
        table["extra_values"]["total_time_sec"]
    )

    write_table_row(table, item)


def _print_console_header(column_specs):
    header = ""
    for column_spec in column_specs:
        if column_spec.console:
            header += ("{:%s%d} " % (column_spec.align, column_spec.width)).format(
                column_spec.header
            )

    print(header)


def print_console_underline(column_specs):
    underline = ""
    for column_spec in column_specs:
        if column_spec.console:
            underline += "-" * column_spec.width + " "

    print(underline)


def _print_csv_header(csv_writer, column_specs):
    csv_writer.writerow([column_spec.header for column_spec in column_specs])


def _transform_if_necessary(column_spec, value, item):
    if column_spec.transform:
        transform = column_spec.transform
        value = transform(value, item)

    if not isinstance(value, str) and not isinstance(value, str):
        value = str(value)

    return value


def _print_console_row(column_specs, item, extra_values=None):
    row = ""

    for column_spec in column_specs:
        if not column_spec.console:
            continue

        value = ""
        if column_spec.field in extra_values:
            value = str(
                extra_values[column_spec.field]
                if extra_values[column_spec.field] is not None
                else ""
            )
        elif column_spec.field:
            value = _get_field_from_obj_or_dict(item, column_spec.field)
            if value is None:
                value = ""

        value = _transform_if_necessary(column_spec, value, item)

        value = value.replace("\n", ", ")

        if len(value) > column_spec.width:
            if column_spec.align == "<":
                value = value[0: column_spec.width - 3] + "..."
            else:
                value = "..." + value[len(value) - column_spec.width + 3:]

        row += ("{:%s%d} " % (column_spec.align, column_spec.width)).format(value)

    # commented out to hide output on data lab
    print(row)


def _get_field_from_obj_or_dict(item, field):
    try:
        value = item[field]
    except TypeError:
        value = getattr(item, field)

    return value


def _print_csv_row(csv_writer, column_specs, item, extra_values):
    columns = []

    for column_spec in column_specs:

        value = ""
        if column_spec.field in extra_values:
            value = str(
                extra_values[column_spec.field]
                if extra_values[column_spec.field] is not None
                else ""
            )
        elif column_spec.field:
            value = _get_field_from_obj_or_dict(item, column_spec.field)
            if value is None:
                value = ""

        value = _transform_if_necessary(column_spec, value, item)

        if not value:
            value = ""

        columns.append(value)

    csv_writer.writerow([s for s in columns])



from __future__ import annotations

import copy
import json
import os
import re
import textwrap
from typing import List, Dict, Union, Optional, Tuple

import pandas as pd

from seeq.base import util
from seeq.sdk import *
from seeq.spy import _common, _metadata
from seeq.spy._errors import *
from seeq.spy._redaction import safely
from seeq.spy._session import Session
from seeq.spy._status import Status
from seeq.spy.workbooks._context import WorkbookPushContext
from seeq.spy.workbooks._item import Item, ItemMap, ItemExists


class StoredOrCalculatedItem(Item):
    def __init__(self, definition=None, *, provenance=None):
        super().__init__(definition, provenance=provenance)

    def _populate_asset_and_path(self, session: Session, item_id, status, item_search_preview: ItemSearchPreviewV1):
        if item_search_preview is None:
            trees_api = TreesApi(session.client)
            asset_tree_output = safely(
                lambda: trees_api.get_tree(id=item_id),
                action_description=f'get Tree information for {item_id}',
                status=status)  # type: AssetTreeOutputV1
            if asset_tree_output is None or asset_tree_output.item is None:
                return

            ancestors = [a.name for a in asset_tree_output.item.ancestors]
            if len(asset_tree_output.item.ancestors) > 0:
                self.definition['Parent ID'] = asset_tree_output.item.ancestors[-1].id
        else:
            ancestors = [a.name for a in item_search_preview.ancestors]
            if len(item_search_preview.ancestors) > 0:
                self.definition['Parent ID'] = item_search_preview.ancestors[-1].id

        _common.add_ancestors_to_definition(self.type, self.name, ancestors, self.definition, old_asset_format=False)

    def _pull(self, session: Session, item_id, status: Status, item_search_preview: ItemSearchPreviewV1):
        super()._pull(session, item_id, status, item_search_preview)

        if 'Path' not in self.definition:
            self._populate_asset_and_path(session, item_id, status, item_search_preview)

    def get_metadata_to_push(self, context: WorkbookPushContext, datasource_maps, datasource_output,
                             item_inventory: dict, *, pushed_workbook_id=None, item_map: ItemMap = None, label=None,
                             item_exists: ItemExists = ItemExists.MAYBE,
                             item_search_preview: ItemSearchPreviewV1 = None):
        raise SPyRuntimeError('get_metadata_to_push() called but StoredOrCalculatedItem.get_metadata_to_push() not '
                              'overloaded')

    def _attach_to_parent(self, session: Session, item_map: ItemMap, item_id: str):
        if self['Scoped To'] is None:
            # Do not move around global items
            return

        if 'Parent ID' not in self.definition:
            return

        parent_id = self.definition['Parent ID']
        if parent_id != Item.ROOT and self.definition['Parent ID'] not in item_map:
            return

        trees_api = TreesApi(session.client)
        if parent_id == Item.ROOT:
            trees_api.move_nodes_to_root_of_tree(body=ItemIdListInputV1(items=[item_id]))
        else:
            trees_api.move_nodes_to_parent(parent_id=item_map[self.definition['Parent ID']],
                                           body=ItemIdListInputV1(items=[item_id]))

    @staticmethod
    def _find_datasource_name(datasource_class, datasource_id, datasource_maps):
        for datasource_map in datasource_maps:
            if datasource_map['Datasource Class'] == datasource_class and \
                    datasource_map['Datasource ID'] == datasource_id:
                return datasource_map['Datasource Name']

        raise SPyRuntimeError('Could not find Datasource Class "%s" and Datasource ID "%s" in datasource maps' %
                              (datasource_class, datasource_id))

    def _execute_regex_map(self, old_definition, regex_map, regex_map_index, item_map: ItemMap, *,
                           allow_missing_properties=False):
        for prop in ['Old', 'New']:
            if prop not in regex_map:
                raise SPyValueError(f'RegEx-Based Map {regex_map_index} must have "{prop}" property')

        capture_groups = dict()
        for prop, regex in regex_map['Old'].items():
            if prop not in StoredItem.SEARCHABLE_PROPS:
                searchable_props_string = '\n'.join(StoredItem.SEARCHABLE_PROPS)
                raise SPyRuntimeError(
                    f'Datasource Map contains an unsearchable property "{prop}".'
                    f'Searchable properties:\n{searchable_props_string}')

            if prop not in old_definition:
                if allow_missing_properties:
                    continue
                else:
                    item_map.log(self.id, f'- RegEx-Based Map {regex_map_index}: Unsuccessful match. Details:')
                    item_map.log(self.id, f'    "{prop}" not defined for this item')
                    return None, None

            pythonized_regex = util.pythonize_regex_capture_group_names(regex)
            match = re.fullmatch(pythonized_regex, old_definition[prop])
            if not match:
                item_map.log(self.id, f'- RegEx-Based Map {regex_map_index}: Unsuccessful match. Details:')
                item_map.log(self.id, f'    "{prop}"')
                item_map.log(self.id, f'        regex          "{regex}"')
                item_map.log(self.id, f'        does not match "{old_definition[prop]}"')
                return None, None

            capture_groups.update(match.groupdict())

        new_definition = dict()
        for prop, regex in regex_map['New'].items():
            new_definition[prop] = util.replace_tokens_in_regex(regex, capture_groups, escape=False)

        return new_definition, capture_groups

    def _lookup_item_via_datasource_map(
            self, session: Session, pushed_workbook_id: str, datasource_maps: List[Dict],
            item_map: ItemMap, *, only_override_maps: bool = False, force_direct_lookup: bool = False,
            item_exists: ItemExists = ItemExists.MAYBE,
            item_search_preview: Optional[ItemSearchPreviewV1] = None) -> Optional[Item]:
        item: Optional[Item] = None

        item_map.clear_logs(self.id)

        # First, we process the "overrides". These are the cases where, even if the item with the ID exists in the
        # destination, we still want to map to something else. Useful for swapping datasources on the same server.
        overrides = [m for m in datasource_maps if _common.get(m, 'Override', default=False)]
        if len(overrides) > 0:
            if "File" in overrides[0]:
                item_map.log(self.id, f'Using overrides from {os.path.dirname(overrides[0]["File"])}:')
            item = self._lookup_in_datasource_map(session, pushed_workbook_id, overrides, item_map)
        else:
            item_map.log(self.id, f'No datasource map overrides found')

        if item is not None:
            item_map.log(self.id, f'Successful mapping:')
            item_map.log(self.id, f'  Old: {self}')
            item_map.log(self.id, f'  New: {item}')
            return item

        if only_override_maps:
            item_map.log(self.id, f'{self} not mapped, only override maps used', at_top=True)
            if not force_direct_lookup:
                return None

        # Second, we just try to look the item up by its ID. This case will occur when the user pulls a workbook,
        # makes a change, and pushes it back.
        if item_exists != ItemExists.NO:
            try:
                item = Item.pull(self.id, session=session, item_search_preview=item_search_preview)
            except ApiException:
                item_map.log(self.id, f"Item's ID {self.id} not found directly in target server")
        else:
            item_map.log(self.id, f"Item's ID {self.id} not found directly in target server")

        if item is not None:
            item_map.log(self.id, f"Item's ID {self.id} found directly in target server")
            return item

        if only_override_maps:
            return None

        # Finally, we try to use the non-override maps to find the item. This case will occur mostly when
        # transferring workbooks between servers.
        non_overrides = [m for m in datasource_maps if not _common.get(m, 'Override', default=False)]
        if len(non_overrides) > 0:
            if "File" in non_overrides[0]:
                item_map.log(self.id, f'Using non-overrides from {os.path.dirname(non_overrides[0]["File"])}:')
            item = self._lookup_in_datasource_map(session, pushed_workbook_id, non_overrides, item_map)
        else:
            item_map.log(self.id, f'No (non-override) datasource maps found')

        if item is None:
            item_map.log(self.id, f'{self} not successfully mapped', at_top=True)
            return None

        item_map.log(self.id, f'Successful mapping:')
        item_map.log(self.id, f'  Old: {self}')
        item_map.log(self.id, f'  New: {item}')

        return item

    def _lookup_in_datasource_map(self, session: Session, pushed_workbook_id, datasource_maps,
                                  item_map: ItemMap) -> Optional[Item]:
        item: Optional[Item] = None

        relevant_maps = list()
        for datasource_map in datasource_maps:
            if _common.get(self, 'Datasource Class') != _common.get(datasource_map, 'Datasource Class') or \
                    _common.get(self, 'Datasource ID') != _common.get(datasource_map, 'Datasource ID'):
                continue

            relevant_maps.append(datasource_map)
            if "File" in datasource_map:
                item_map.log(self.id, f'- Used "{datasource_map["File"]}"')
            else:
                item_map.log(self.id,
                             f'- Used datasource map for Datasource Class "{datasource_map["Datasource Class"]}" '
                             f'and Datasource ID "{datasource_map["Datasource ID"]}"')

            if _common.DATASOURCE_MAP_ITEM_LEVEL_MAP_FILES in datasource_map:
                item = self._lookup_in_item_level_map_files(session, datasource_map, item_map)

            if item is not None:
                break

            if _common.DATASOURCE_MAP_REGEX_BASED_MAPS in datasource_map:
                item = self._lookup_in_regex_based_maps(session, datasource_map, datasource_maps, item_map,
                                                        pushed_workbook_id)

            if item is not None:
                break

        if len(relevant_maps) == 0:
            item_map.log(
                self.id,
                f'- No Datasource Maps found that match Datasource Class "{_common.get(self, "Datasource Class")}" '
                f'and Datasource ID "{_common.get(self, "Datasource ID")}"')

        return item

    def _lookup_in_item_level_map_files(self, session: Session, datasource_map,
                                        item_map: ItemMap) -> Optional[Item]:
        item: Optional[ItemOutputV1] = None
        item_level_map_files = datasource_map['Item-Level Map Files']
        if not isinstance(item_level_map_files, list):
            item_level_map_files = [item_level_map_files]

        if 'DataFrames' not in datasource_map:
            datasource_map['DataFrames'] = dict()

        df_map = datasource_map['DataFrames']

        for item_level_map_file in item_level_map_files:
            if item_level_map_file not in df_map:
                if not util.safe_exists(item_level_map_file):
                    raise SPyValueError('Item-Level Map File "%s" does not exist' % item_level_map_file)

                df_map[item_level_map_file] = pd.read_csv(item_level_map_file)

            df = df_map[item_level_map_file]

            if 'Old ID' not in df.columns or 'New ID' not in df.columns:
                raise SPyValueError('Item-Level Map File must have "Old ID" and "New ID" columns')

            mapped_row = df[df['Old ID'] == self.id]

            if len(mapped_row) > 1:
                raise SPyValueError('Ambiguous map: %d rows in "%s" match "Old ID" == "%s"\n%s' % (
                    len(mapped_row), item_level_map_file, self.id, mapped_row))

            if len(mapped_row) == 0:
                item_map.log(
                    self.id,
                    f'- Item-Level Map File "{item_level_map_file}" did not match for "Old ID" value "{self.id}"')
                continue

            new_id = mapped_row.iloc[0]['New ID']

            item = Item.pull(new_id, session=session)
            break

        return item

    def _lookup_in_regex_based_maps(
            self, session: Session, datasource_map, datasource_maps, item_map: ItemMap,
            pushed_workbook_id) -> Optional[Union[ItemOutputV1, UserOutputV1, IdentityPreviewV1]]:

        for i in range(len(datasource_map['RegEx-Based Maps'])):
            regex_map_index = i
            regex_map = datasource_map['RegEx-Based Maps'][regex_map_index]
            item_object, old_criteria_matched = self._lookup_in_regex_based_map(
                session, regex_map_index, regex_map, datasource_maps, item_map, pushed_workbook_id)

            if item_object is not None:
                return item_object

            on_match_default = 'Stop' if session.options.wants_compatibility_with(191) else 'Continue'
            on_match = str(_common.get(regex_map, 'On Match', on_match_default)).lower()
            if not isinstance(on_match, str) and on_match not in ['continue', 'stop']:
                raise SPyValueError('RegEx-Based Map "On Match" parameter must be either "Continue" or "Stop"')

            if old_criteria_matched and on_match == 'stop':
                item_map.log(self.id,
                             f'- "On Match" parameter is "Stop", so no further RegEx-Based Maps will be used')
                break

        return None

    def _lookup_in_regex_based_map(
            self, session: Session, regex_map_index, regex_map, datasource_maps, item_map: ItemMap,
            pushed_workbook_id) -> Tuple[Optional[Union[ItemOutputV1, UserOutputV1, IdentityPreviewV1]], bool]:
        items_api = ItemsApi(session.client)
        old_definition = self.definition_dict
        new_definition = None
        item: Optional[Item] = None
        capture_groups = None

        if 'Type' not in old_definition:
            raise SPyValueError('"Type" property required in "Old" datasource map definitions')

        if 'Datasource Class' in old_definition and 'Datasource ID' in old_definition:
            old_definition['Datasource Name'] = \
                StoredOrCalculatedItem._find_datasource_name(old_definition['Datasource Class'],
                                                             old_definition['Datasource ID'],
                                                             datasource_maps)

        new_definition, capture_groups = self._execute_regex_map(
            old_definition, regex_map, regex_map_index, item_map,
            allow_missing_properties=self.type.endswith('Datasource'))

        if new_definition is None:
            return None, False

        if 'Type' not in new_definition:
            raise SPyValueError('"Type" property required in "New" datasource map definitions')

        if 'Datasource Name' in new_definition:
            if 'Datasource ID' not in new_definition:
                if 'Datasource Class' not in new_definition:
                    raise SPyValueError('"Datasource Class" required with "Datasource Name" in map:\n%s' %
                                        json.dumps(new_definition))

                datasource_results = items_api.search_items(
                    # Note here that 'Name' supports RegEx but 'Datasource Class' doesn't. Users have to be careful in
                    # the Datasource Maps not to supply a RegEx. See Workbook._construct_default_datasource_maps().
                    filters=['Datasource Class==%s&&Name==/%s/' % (new_definition['Datasource Class'],
                                                                   new_definition['Datasource Name']),
                             '@includeUnsearchable'],
                    types=['Datasource'],
                    limit=2)  # type: ItemSearchPreviewPaginatedListV1

                items = datasource_results.items
                if len(items) > 1:
                    # Try filtering out any archived items first
                    items = [i for i in items if not i.is_archived]

                if len(items) > 1:
                    multiple_items_str = "\n".join([i.id for i in items])
                    raise SPyRuntimeError(
                        f'Multiple datasources found that match "{new_definition["Datasource Name"]}":\n'
                        f'{multiple_items_str}'
                    )
                elif len(items) == 0:
                    item_map.log(self.id,
                                 f'- No datasource found that matches "{new_definition["Datasource Name"]}"')
                    return None, True

                new_datasource = items[0]  # type: ItemSearchPreviewV1
                try:
                    new_definition['Datasource ID'] = items_api.get_property(
                        id=new_datasource.id, property_name='Datasource ID').value
                except ApiException:
                    raise SPyRuntimeError(f'Could not get Datasource ID for '
                                          f'"{new_definition["Datasource Name"]}" {new_datasource.id}')

            del new_definition['Datasource Name']

        new_type = _common.get(new_definition, 'Type', '?')
        # Simplify the types so that the search will find stored or calculated signals/conditions etc. This
        # helps when mapping, for example, an AF tree to a Data Lab tree.
        for t in ['Signal', 'Condition', 'Scalar', 'Datasource']:
            if t in new_type:
                new_definition['Type'] = t

        if new_type not in ['User', 'UserGroup']:
            query_df = pd.DataFrame([new_definition])
            from seeq.spy import _search
            if 'Path' in new_definition:
                # If we're matching based on path, unfortunately we will not be able to match against archived items
                # due to a limitation of the API. See CRAB-18439.
                search_df = _search.search(query_df, workbook=pushed_workbook_id, recursive=False,
                                           ignore_unindexed_properties=False, session=session, quiet=True)
            else:
                search_df = _search.search(query_df, workbook=pushed_workbook_id, include_archived=True,
                                           ignore_unindexed_properties=False, session=session, quiet=True)

            # If several things are returned, but some of them are archived, then filter out the archived ones.
            # Tested by spy.workbooks.tests.test_push.test_datasource_map_by_name()
            if len(search_df) > 1 and 'Archived' in search_df.columns:
                search_df = search_df[~search_df['Archived']]

            def _log_criteria(_and_found=None):
                for key, value in new_definition.items():
                    item_map.log(self.id, f'    "{key}"')
                    if key in regex_map['Old']:
                        item_map.log(self.id, f'        regex          "{regex_map["Old"][key]}"')
                        item_map.log(self.id, f'        matched on     "{self[key]}"')
                    item_map.log(self.id, f'        searched for   "{value}"')
                    if _and_found is not None:
                        item_map.log(self.id, f'        and found      "{_and_found[key]}"')
                if capture_groups is not None and len(capture_groups) > 0:
                    item_map.log(self.id, f'    Capture groups:')
                    for n, v in capture_groups.items():
                        item_map.log(self.id, f'        {n:14} "{v}"')

            if len(search_df) == 0:
                item_map.log(self.id, f'- RegEx-Based Map {regex_map_index}: Item not found on server. Details:')
                _log_criteria()
            elif len(search_df) > 1:
                item_map.log(self.id,
                             f'- RegEx-Based Map {regex_map_index}: Multiple possibilities for item found. Details:')
                _log_criteria()
                item_map.log(self.id, f'  Matching items:')
                item_map.log(self.id, textwrap.indent(f'{str(search_df[["ID", "Name"]])}', '    '))
            else:
                item = Item.pull(search_df.iloc[0]['ID'], session=session)
                item_map.log(self.id, f'- RegEx-Based Map {regex_map_index}: Successfully mapped. Details:')
                _log_criteria(item)
        else:
            if new_type == 'User':
                try:
                    users_api = UsersApi(session.client)
                    user = users_api.get_user_from_username(
                        auth_datasource_class=new_definition['Datasource Class'],
                        auth_datasource_id=new_definition['Datasource ID'],
                        username=new_definition['Username'])
                    from seeq.spy.workbooks._user import User
                    item = User.pull(user.id, session=session)
                except ApiException:
                    # Fall through, item not found
                    pass
            else:
                users_api = UsersApi(session.client)
                offset = 0
                limit = 100

                while True:
                    identity_preview_list = users_api.autocomplete_users_and_groups(query=new_definition['Name'],
                                                                                    offset=offset,
                                                                                    limit=limit)

                    for identity in identity_preview_list.items:  # type: IdentityPreviewV1
                        if identity.type == 'UserGroup' and identity.name == new_definition['Name']:
                            from seeq.spy.workbooks._user import UserGroup
                            item = UserGroup.from_identity(identity, session=session)
                            break

                    if item is not None or len(identity_preview_list.items) < limit:
                        break

                    offset += limit

        return item, True


class PushThis(Exception):
    def __init__(self, definition):
        self.definition = definition


class StoredItem(StoredOrCalculatedItem):
    SEARCHABLE_PROPS = ['Datasource Class', 'Datasource ID', 'Datasource Name', 'Data ID',
                        'Type', 'Name', 'Description', 'Username', 'Path', 'Asset']

    def get_metadata_to_push(self, context: WorkbookPushContext, datasource_maps, datasource_output,
                             item_inventory: dict, *, pushed_workbook_id=None, item_map: ItemMap = None, label=None,
                             item_exists: ItemExists = ItemExists.MAYBE,
                             item_search_preview: ItemSearchPreviewV1 = None):
        try:
            self._lookup(context.session, datasource_maps, datasource_output,
                         pushed_workbook_id=pushed_workbook_id, item_map=item_map, label=label,
                         override_max_interp=context.override_max_interp,
                         dummy_items_workbook_context=context.dummy_items_workbook_context,
                         item_exists=item_exists, item_search_preview=item_search_preview, context=context,
                         item_inventory=item_inventory)
        except PushThis as e:
            return e.definition

        return None

    def _push_dummy_item(self, context: WorkbookPushContext, label: str, item_map: ItemMap,
                         datasource_output: DatasourceOutputV1, dummy_items_workbook_context):
        metadata_dict = self.definition_dict.copy()

        if context.global_inventory == 'always reuse' and not _common.present(self.definition, 'Scoped To'):
            label = None

        metadata_dict['Dummy Item'] = True

        metadata_dict['Original ID'] = metadata_dict['ID']
        metadata_dict['Original Datasource Class'] = metadata_dict['Datasource Class']
        metadata_dict['Original Datasource ID'] = metadata_dict['Datasource ID']
        metadata_dict['Original Data ID'] = metadata_dict['Data ID']

        del metadata_dict['ID']
        metadata_dict['Datasource Class'] = datasource_output.datasource_class
        metadata_dict['Datasource ID'] = datasource_output.datasource_id
        metadata_dict['Data ID'] = self._construct_data_id(
            label, for_item=self, reconcile_by=context.reconcile_inventory_by,
            workbook_id=dummy_items_workbook_context.workbook_object.id)

        if 'Parent ID' in metadata_dict:
            # Dummy items will be pushed with a Path and Asset so that it gets put in a (minimal) tree and therefore can
            # be asset-swapped.
            del metadata_dict['Parent ID']

        # Scope it to the dummy item workbook
        metadata_dict['Scoped To'] = dummy_items_workbook_context.workbook_object.id

        item_map.log(self.id, f'Dummy item pushed: {self}')

        raise PushThis(metadata_dict)

    def _push_local_item(self, context: WorkbookPushContext, label: str, item_map: ItemMap,
                         datasource_output: DatasourceOutputV1, pushed_workbook_id: str,
                         item_inventory: dict, item_exists: bool = False):
        metadata_dict = self.definition_dict.copy()

        metadata_dict['Dummy Item'] = (metadata_dict.get('Type') != 'Asset')

        if 'Parent ID' in metadata_dict and metadata_dict['Parent ID'] == Item.ROOT:
            del metadata_dict['Parent ID']

        if not item_exists:
            del metadata_dict['ID']
            metadata_dict['Datasource Class'] = datasource_output.datasource_class
            metadata_dict['Datasource ID'] = datasource_output.datasource_id
            metadata_dict['Data ID'] = self._construct_data_id(
                label, for_item=self, reconcile_by=context.reconcile_inventory_by,
                workbook_id=pushed_workbook_id)

            if 'Parent ID' in metadata_dict:
                metadata_dict['Parent Data ID'] = self._construct_data_id(
                    label, for_item=item_inventory[self.definition['Parent ID']],
                    reconcile_by=context.reconcile_inventory_by, workbook_id=pushed_workbook_id)
                del metadata_dict['Parent ID']

        # Scope it to the pushed workbook
        metadata_dict['Scoped To'] = pushed_workbook_id

        item_map.log(self.id, f'Local item pushed: {self}')

        raise PushThis(metadata_dict)

    def _lookup(self, session: Session, datasource_maps, datasource_output, *,
                pushed_workbook_id=None, item_map: ItemMap = None, label=None, override_max_interp=False,
                dummy_items_workbook_context=None, item_exists: ItemExists = ItemExists.MAYBE,
                item_search_preview: Optional[ItemSearchPreviewV1] = None,
                context: Optional[WorkbookPushContext] = None, item_inventory: Optional[dict] = None):
        # Note: The caller of this function is in charge of API safety

        if self.id in item_map.data_item_cache:
            return item_map.data_item_cache[self.id]

        local: bool = self['Scoped To'] is not None
        only_override_maps = item_map.only_override_maps
        if local:
            only_override_maps = True

        force_direct_lookup = label is None

        item = self._lookup_item_via_datasource_map(session, pushed_workbook_id, datasource_maps, item_map,
                                                    only_override_maps=only_override_maps,
                                                    force_direct_lookup=force_direct_lookup,
                                                    item_exists=item_exists,
                                                    item_search_preview=item_search_preview)

        if item is None:
            if local:
                self._push_local_item(context, label, item_map, datasource_output, pushed_workbook_id, item_inventory)
            else:
                if dummy_items_workbook_context is None:
                    raise SPyDependencyNotFound(item_map.explain(self.id))

                self._push_dummy_item(context, label, item_map, datasource_output, dummy_items_workbook_context)
        else:
            if local:
                item_scope = _common.get(item, 'Scoped To')
                item_is_scoped_to_other_workbook = item_scope is not None and item_scope != pushed_workbook_id
                self_already_exists_and_may_need_update = (
                        item_scope is not None
                        and item_scope == pushed_workbook_id
                        and _common.get(self, 'ID') == _common.get(item, 'ID')
                        and not label
                )
                if item_is_scoped_to_other_workbook or self_already_exists_and_may_need_update:
                    self._push_local_item(
                        context, label, item_map, datasource_output, pushed_workbook_id, item_inventory,
                        self_already_exists_and_may_need_update)

        item_map[self.id] = item.id
        item_map.data_item_cache[self.id] = item

        if item.type not in ['User', 'UserGroup']:
            # We need to exclude Example Data, because it is set explicitly by the connector
            datasource_class = item['Datasource Class']
            datasource_id = item['Datasource ID']
            is_example_data = (datasource_class and datasource_class == 'Time Series CSV Files' and
                               datasource_id and datasource_id == 'Example Data')

            if override_max_interp and item.type == 'StoredSignal' and not is_example_data and \
                    'Maximum Interpolation' in self:
                src_max_interp = Item._property_input_from_scalar_str(self['Maximum Interpolation'])
                dst_max_interp_prop = item['Maximum Interpolation']
                if (dst_max_interp_prop and hasattr(dst_max_interp_prop, 'value')
                        and hasattr(dst_max_interp_prop, 'unit_of_measure')):
                    dst_max_interp = Item._property_input_from_scalar_str(dst_max_interp_prop.value)
                    if src_max_interp.unit_of_measure != dst_max_interp.unit_of_measure or \
                            src_max_interp.value != dst_max_interp.value:
                        items_api = ItemsApi(session.client)
                        items_api.set_property(id=item.id,
                                               property_name='Override Maximum Interpolation',
                                               body=src_max_interp)

        return item


class Asset(StoredItem):
    def __init__(self, definition=None, *, provenance=None):
        super().__init__(definition, provenance=provenance)

        if self.provenance == Item.LOAD:
            if self.definition.get('Old Asset Format', True):
                if self.type == 'Asset':
                    if self.definition.get('Path') in [None, '']:
                        self.definition['Path'] = self.definition['Asset']
                    else:
                        self.definition['Path'] += ' >> ' + self.definition['Asset']
                        self.definition['Asset'] = self.definition['Name']

        self.definition['Old Asset Format'] = False


class Datasource(StoredItem):
    """
    The SPy representation of a Datasource, the 'source' container for a collection of other items.
    """

    @staticmethod
    def from_datasource_output(datasource_output: Union[DatasourceOutputV1, DatasourcePreviewV1]):
        return Datasource({
            'ID': datasource_output.id,
            'Name': datasource_output.name,
            'Datasource Class': datasource_output.datasource_class,
            'Datasource ID': datasource_output.datasource_id,
            'Archived': datasource_output.is_archived
        })


class CalculatedItem(StoredOrCalculatedItem):
    def _lookup(self, session: Session, datasource_maps, pushed_workbook_id, item_map: ItemMap,
                label: str, item_exists: ItemExists, item_search_preview: Optional[ItemSearchPreviewV1]):
        if self.id in item_map.data_item_cache:
            return item_map.data_item_cache[self.id]

        # This is a key piece of datasource mapping logic. CalculatedItems that are in trees are more like "source
        # data" in the context of workbooks/worksheets. In other words, when we push them, we want to map in existing
        # items rather than creating a standalone CalculatedItem that isn't a part of a tree, has no associated asset
        # and cannot be swapped.
        is_standalone_item = _common.get(self, 'Asset') is None

        # So if it's in a global tree, we process the datasource maps just like we would if it were a StoredItem. If
        # it's *not* in a global tree, then we only process datasource maps that have been supplied in a
        # datasource_map_folder and are therefore treated as "overrides". CalculatedItems that are associated with the
        # particular analysis of a workbook should be created/overwritten as part of the push, and since they won't
        # exist in a tree, we purposefully don't attempt to do any mapping and we fall through to the CalculatedItem's
        # create/update code.
        local: bool = self['Scoped To'] is not None
        only_override_maps = item_map.only_override_maps
        if local or is_standalone_item:
            only_override_maps = True

        item = self._lookup_item_via_datasource_map(
            session, pushed_workbook_id, datasource_maps, item_map, only_override_maps=only_override_maps,
            item_exists=item_exists, item_search_preview=item_search_preview)

        if not local and not is_standalone_item and item is None:
            raise SPyDependencyNotFound(item_map.explain(self.id))

        if item:
            item_map[self.id] = item.id
            item_map.data_item_cache[self.id] = item

        return item

    def _find_by_name(self, session: Session, workbook_id):
        return Item.find_item_by_search(session, f'Name=={self.name}', scope=[workbook_id], types=[self.type])

    def get_metadata_to_push(self, context: WorkbookPushContext, datasource_maps, datasource_output,
                             item_inventory: dict, *, pushed_workbook_id=None, item_map: ItemMap = None, label=None,
                             item_exists: ItemExists = ItemExists.MAYBE,
                             item_search_preview: ItemSearchPreviewV1 = None):
        item = self._lookup(context.session, datasource_maps, pushed_workbook_id, item_map, label, item_exists,
                            item_search_preview)

        if item:
            return None

        item_output = item_search_preview
        if self.provenance == Item.CONSTRUCTOR:
            if item_exists == ItemExists.MAYBE:
                item_output = self.find_me(context.session, label, datasource_output)

            if item_output is None:
                item_output = self._find_by_name(context.session, pushed_workbook_id)

        if context.global_inventory == 'always reuse' and not _common.present(self.definition, 'Scoped To'):
            label = None

        if label is None and item_exists != ItemExists.NO and item_search_preview is None:
            item_output = self.find_me(context.session, label, datasource_output)

        metadata_dict = copy.deepcopy(self.definition_dict)

        if 'Formula' in self.definition:
            metadata_dict['Formula'] = Item.formula_string_from_list(self.definition['Formula'])

        if context.global_inventory == 'copy local' or _common.present(self.definition, 'Scoped To'):
            metadata_dict['Scoped To'] = pushed_workbook_id

        if 'Scoped To' not in metadata_dict:
            pushed_workbook_id = None

        # Custody ID preserves the original ID of the item which helps maintain a consistent Data ID for the same
        # item through multiple round-trips.
        metadata_dict['Custody ID'] = self.get('Custody ID', self.id)

        if (label is None or self.provenance == Item.CONSTRUCTOR) and item_output is not None:
            def _prop(_n):
                if isinstance(item_output, ItemSearchPreviewV1):
                    return item_output.included_properties[_n].value
                elif isinstance(item_output, ItemOutputV1):
                    return [_p.value for _p in item_output.properties if _p.name == _n][0]

            metadata_dict['Datasource Class'] = _prop('Datasource Class')
            metadata_dict['Datasource ID'] = _prop('Datasource ID')
            metadata_dict['Data ID'] = _prop('Data ID')

            if (metadata_dict['Datasource Class'] == datasource_output.datasource_class and
                    metadata_dict['Datasource ID'] == datasource_output.datasource_id):
                # This allows the batch endpoints to be used
                del metadata_dict['ID']

                if 'Parent ID' in metadata_dict:
                    metadata_dict['Parent Data ID'] = self._construct_data_id(
                        label, for_item=item_inventory[self.definition['Parent ID']],
                        reconcile_by=context.reconcile_inventory_by, workbook_id=pushed_workbook_id)
        else:
            del metadata_dict['ID']
            metadata_dict['Datasource Class'] = datasource_output.datasource_class
            metadata_dict['Datasource ID'] = datasource_output.datasource_id
            metadata_dict['Data ID'] = self._construct_data_id(
                label, for_item=self, reconcile_by=context.reconcile_inventory_by, workbook_id=pushed_workbook_id)

            if 'Parent ID' in metadata_dict:
                metadata_dict['Parent Data ID'] = self._construct_data_id(
                    label, for_item=item_inventory[self.definition['Parent ID']],
                    reconcile_by=context.reconcile_inventory_by, workbook_id=pushed_workbook_id)

        return metadata_dict

    def _set_ui_config(self, session: Session, _id):
        if 'UIConfig' in self.definition:
            try:
                items_api = ItemsApi(session.client)
                items_api.set_property(id=_id, property_name='UIConfig',
                                       body=PropertyInputV1(value=json.dumps(self.definition['UIConfig'])))
            except ValueError:  # ValueError includes JSONDecodeError, but ApiException will not be caught
                pass


class StoredSignal(StoredItem):
    """
    The SPy representation of a Stored Signal, a collection of samples composed of timestamp + value pairs.
    These samples may be stored in an external datasource or within the Seeq database.
    """
    pass


class CalculatedSignal(CalculatedItem):
    """
    The SPy representation of a Calculated Signal, a collection of samples composed of timestamp + value pairs.
    These samples are calculated based on formula inputs.
    """

    def _pull(self, session: Session, item_id, status, item_search_preview: ItemSearchPreviewV1):
        super()._pull(session, item_id, status, item_search_preview)

        if item_search_preview is None:
            # Note: The caller of this function is in charge of API safety
            signals_api = SignalsApi(session.client)
            signal_output = signals_api.get_signal(id=item_id)  # type: SignalOutputV1
            self._set_formula_based_item_properties(signal_output.parameters)
        else:
            self._set_formula_based_item_properties(item_search_preview.parameters)


class StoredCondition(StoredItem):
    """
    The SPy representation of a Stored Condition, a collection of capsules composed of start & end timestamps
    plus optional properties.
    These capsules may be stored in an external datasource or within the Seeq database.
    """
    pass


class CalculatedCondition(CalculatedItem):
    """
    The SPy representation of a Calculated Condition, a collection of capsules composed of start & end timestamps
    plus optional properties.
    These capsules are calculated based on formula inputs.
    """

    def _pull(self, session: Session, item_id, status, item_search_preview: ItemSearchPreviewV1):
        super()._pull(session, item_id, status, item_search_preview)

        if item_search_preview is None:
            # Note: The caller of this function is in charge of API safety
            conditions_api = ConditionsApi(session.client)
            condition_output = conditions_api.get_condition(id=item_id)  # type: ConditionOutputV1
            self._set_formula_based_item_properties(condition_output.parameters)
        else:
            self._set_formula_based_item_properties(item_search_preview.parameters)


class LiteralScalar(StoredItem):
    """
    The SPy representation of a Literal Scalar, a single value with an optional unit of measure.
    This value may be stored in an external datasource or within the Seeq database.
    """
    pass


class CalculatedScalar(CalculatedItem):
    """
    The SPy representation of a Calculated Scalar, a single value with an optional unit of measure.
    This value is calculated based on formula inputs.
    """

    def _pull(self, session: Session, item_id, status: Status, item_search_preview: ItemSearchPreviewV1):
        super()._pull(session, item_id, status, item_search_preview)
        if item_search_preview is None:
            # Note: The caller of this function is in charge of API safety
            scalars_api = ScalarsApi(session.client)
            calculated_item_output = scalars_api.get_scalar(id=item_id)  # type: CalculatedItemOutputV1
            self._set_formula_based_item_properties(calculated_item_output.parameters)
        else:
            self._set_formula_based_item_properties(item_search_preview.parameters)


class Chart(CalculatedItem):
    """
    The SPy representation of a Chart or Histogram, a calculation that accepts a time range and computes
    aggregated values within this time range.
    The values are calculated based on formula inputs and the time range.
    """

    def _pull(self, session: Session, item_id, status: Status, item_search_preview: ItemSearchPreviewV1):
        super()._pull(session, item_id, status, item_search_preview)

        if item_search_preview is None:
            # Note: The caller of this function is in charge of API safety
            formulas_api = FormulasApi(session.client)
            calculated_item_output = formulas_api.get_function(id=item_id)  # type: CalculatedItemOutputV1

            self._set_formula_based_item_properties(calculated_item_output.parameters)
        else:
            self._set_formula_based_item_properties(item_search_preview.parameters)

        if 'FormulaParameters' in self.definition:
            # Charts have these superfluous properties
            del self.definition['FormulaParameters']


class ThresholdMetric(CalculatedItem):
    """
    The SPy representation of a Threshold Metric, a table-like calculation aggregates values and can have threshold
    states based on those aggregated values.
    The values are calculated based on formula inputs.
    """

    def _pull(self, session: Session, item_id, status: Status, item_search_preview: ItemSearchPreviewV1):
        super()._pull(session, item_id, status, item_search_preview)
        # Note: The caller of this function is in charge of API safety
        formula_parameters = _metadata.formula_parameters_dict_from_threshold_metric(session, item_id)

        # These properties come through in the GET /items/{id} call, and for clarity's sake we remove them
        for ugly_duplicate_property in ['AggregationFunction', 'BoundingConditionMaximumDuration']:
            if ugly_duplicate_property in self.definition:
                del self.definition[ugly_duplicate_property]

        self.definition['Formula'] = '<ThresholdMetric>'
        self.definition['Formula Parameters'] = formula_parameters
